import torch
from torch import nn
from torch.nn import functional as F

from .utils import _SimpleSegmentationModel

from alchemy_cat.torch_tools import Network

__all__ = []


class DeepLabV3ResNet(_SimpleSegmentationModel):
    """
    Implements DeepLabV3 model from
    `"Rethinking Atrous Convolution for Semantic Image Segmentation"
    <https://arxiv.org/abs/1706.05587>`_.
    """
    def get_param_groups(self) -> dict:
        param_groups = {'backbone_weight': list(self.backbone.parameters()),
                        'head_weight': list(self.head.parameters())}

        return param_groups


class DeepLabHeadV3Plus(Network):
    def __init__(self, in_channels, low_level_channels, num_classes, aspp_dilate=(12, 24, 36)):
        super(DeepLabHeadV3Plus, self).__init__()
        self.project = nn.Sequential(
            nn.Conv2d(low_level_channels, 48, 1, bias=False),
            nn.BatchNorm2d(48),
            nn.ReLU(inplace=True),
        )

        self.aspp = ASPP(in_channels, aspp_dilate)

        self.classifier = nn.Sequential(
            nn.Conv2d(304, 256, 3, padding=1, bias=False),
            nn.BatchNorm2d(256),
            nn.ReLU(inplace=True),
            nn.Conv2d(256, num_classes, 1)
        )
        self._init_weight()

        self.in_channel = in_channels
        self.out_channel = num_classes

    def forward(self, feature):
        low_level_feature = self.project(feature['low_level'])
        output_feature = self.aspp(feature['out'])
        output_feature = F.interpolate(output_feature,
                                       size=low_level_feature.shape[2:], mode='bilinear', align_corners=True)
        return self.classifier(torch.cat([low_level_feature, output_feature], dim=1))

    def _init_weight(self):
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight)
            elif isinstance(m, (nn.BatchNorm2d, nn.GroupNorm)):
                nn.init.constant_(m.weight, 1)
                nn.init.constant_(m.bias, 0)


class DeepLabHead(Network):
    def __init__(self, in_channels, num_classes, aspp_dilate=(12, 24, 36)):
        super(DeepLabHead, self).__init__()

        self.classifier = nn.Sequential(
            ASPP(in_channels, aspp_dilate),
            nn.Conv2d(256, 256, 3, padding=1, bias=False),
            nn.BatchNorm2d(256),
            nn.ReLU(inplace=True),
            nn.Conv2d(256, num_classes, 1)
        )
        self._init_weight()

        self.in_channel = in_channels
        self.out_channel = num_classes

    def forward(self, feature):
        return self.classifier(feature['out'])

    def _init_weight(self):
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight)
            elif isinstance(m, (nn.BatchNorm2d, nn.GroupNorm)):
                nn.init.constant_(m.weight, 1)
                nn.init.constant_(m.bias, 0)


class ASPPConv(nn.Sequential, Network):
    def __init__(self, in_channels, out_channels, dilation):
        modules = [
            nn.Conv2d(in_channels, out_channels, 3, padding=dilation, dilation=dilation, bias=False),
            nn.BatchNorm2d(out_channels),
            nn.ReLU(inplace=True)
        ]
        nn.Sequential.__init__(self, *modules)

        self.in_channel = in_channels
        self.out_channel = out_channels


class ASPPPooling(nn.Sequential, Network):
    def __init__(self, in_channels, out_channels):
        nn.Sequential.__init__(self,
                               nn.AdaptiveAvgPool2d(1),
                               nn.Conv2d(in_channels, out_channels, 1, bias=False),
                               nn.BatchNorm2d(out_channels),
                               nn.ReLU(inplace=True))

        self.in_channel = in_channels
        self.out_channel = out_channels

    def forward(self, x):
        size = x.shape[-2:]
        x = super(ASPPPooling, self).forward(x)
        return F.interpolate(x, size=size, mode='bilinear', align_corners=True)


class ASPP(Network):
    def __init__(self, in_channels, atrous_rates):
        super(ASPP, self).__init__()
        out_channels = 256
        modules = []
        modules.append(nn.Sequential(
            nn.Conv2d(in_channels, out_channels, 1, bias=False),
            nn.BatchNorm2d(out_channels),
            nn.ReLU(inplace=True)))

        rate1, rate2, rate3 = tuple(atrous_rates)
        modules.append(ASPPConv(in_channels, out_channels, rate1))
        modules.append(ASPPConv(in_channels, out_channels, rate2))
        modules.append(ASPPConv(in_channels, out_channels, rate3))
        modules.append(ASPPPooling(in_channels, out_channels))

        self.convs = nn.ModuleList(modules)

        self.project = nn.Sequential(
            nn.Conv2d(5 * out_channels, out_channels, 1, bias=False),
            nn.BatchNorm2d(out_channels),
            nn.ReLU(inplace=True),
            nn.Dropout(0.1), )

        self.in_channel = in_channels
        self.out_channel = out_channels

    def forward(self, x):
        res = []
        for conv in self.convs:
            res.append(conv(x))
        res = torch.cat(res, dim=1)
        return self.project(res)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author  : Xiaobo Yang
@Contact : hal_42@zju.edu.cn
@Time    : 2023/3/22 21:18
@File    : cl_loss.py
@Software: PyCharm
@Desc    : 
"""
import torch
from torch import nn

__all__ = ['MultiLabelCLLoss']


class MultiLabelCLLoss(nn.Module):

    def __init__(self,
                 gamma: float | None = None,
                 reduce: str = 'pos_mean'):
        """将多标签分类的正类视作正样本，负类视作负样本，计算对比损失。

        Args:
            gamma: 相似度的放缩因子。
            reduce: 可以为：pos_mean，对每个正负样本对求平均；sample_mean，先对每个样本的正负样本对求平均，
                再对所有样本求平均。
        """
        super().__init__()
        self.gamma = gamma
        self.reduce = reduce

    def forward(self, S: torch.Tensor, cls_lb: torch.Tensor) -> torch.Tensor:
        """将多标签分类的正类视作正样本，负类视作负样本，计算对比损失。

        Args:
            S: (N, G)的相似度图。
            cls_lb: (N, G)的类别标签。

        Returns:
            多标签对比损失。
        """
        # * 提前计算要用到的索引、数量。
        # ** 计算有效anchor的数量。有限anchor指的是至少有一个正样本和负样本的anchor。
        valid_anchor_mask = torch.any(cls_lb == 0, dim=1) & torch.any(cls_lb == 1, dim=1)
        valid_anchor_num = valid_anchor_mask.sum(dtype=torch.int32)
        if valid_anchor_num == 0:
            return S.mean() * 0
        # ** 计算正负样本掩码。
        neg_mask = (cls_lb == 0) * valid_anchor_mask[:, None]
        pos_mask = (cls_lb == 1) * valid_anchor_mask[:, None]
        # ** 计算正样本的锚序号。
        pi_a = torch.nonzero(pos_mask, as_tuple=True)[0]
        # ** 对每个正样本，计算与之共享anchor的正样本数量。
        pi_I = pos_mask.sum(dim=1, dtype=torch.int32)[pi_a]

        # * 计算CL损失。
        # ** 计算放缩后的相似度。
        S = self.gamma * S if self.gamma is not None else S
        # ** 计算所有相似度的指数。
        e_S = torch.exp(S)
        # * 计算Σje^Snj。
        Sigma_e_Sn_j = (e_S * neg_mask).sum(dim=1)
        # * 得到Spi, e^Spi。
        Spi = S[pos_mask]
        e_Spi = e_S[pos_mask]
        # * 得到每个正样本对应的Σje^Snj。
        pi_Sigma_e_Sn_j = Sigma_e_Sn_j[pi_a]

        # * 对每个正样本，计算对比损失。
        pi_cl_loss = -(Spi - torch.log(e_Spi + pi_Sigma_e_Sn_j))

        # * 按照reduction计算平均损失。
        match self.reduce:
            case 'pos_mean':
                return pi_cl_loss.mean()
            case 'sample_mean':
                pi_cl_loss = pi_cl_loss / pi_I
                return pi_cl_loss.sum() / valid_anchor_num
            case _:
                raise ValueError(f"不支持的{self.reduce=}。")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author  : Xiaobo Yang
@Contact : hal_42@zju.edu.cn
@Time    : 2022/7/31 17:00
@File    : auto_allocate_cuda.py
@Software: PyCharm
@Desc    : 自动分配cuda设备。
"""
import os

__all__ = ['allocate_cuda_by_group_rank']

from .cuda_block import cuda_block_util_no_process_on


def allocate_cuda_by_group_rank(group_rank: int, group_cuda_num: int, block: bool=True,
                                verbosity: bool=True) -> tuple[list[int], dict[str, str]]:
    """依据任务组编号，分配cuda设备。

    Args:
        group_rank: 任务组编号。
        group_cuda_num: 任务需要设备数。
        block: 若为True，则阻塞直至分配到的GPU空闲。
        verbosity: 若为True，阻塞时打印心跳。

    Returns:
        分配给当前任务的cuda设备，以及指定了当前设备的环境变量。
    """
    # * 确定组数。
    cudas = [int(c) for c in os.environ['CUDA_VISIBLE_DEVICES'].split(',')]  # 可用CUDA设备。
    cuda_num = len(cudas)  # 卡数。
    assert cuda_num % group_cuda_num == 0
    group_num = cuda_num // group_cuda_num  # 组数。

    # * 找到当前应当使用的CUDA设备。
    group_idx = group_rank % group_num  # 当前所在组。
    current_cudas = cudas[group_idx * group_cuda_num:(group_idx + 1) * group_cuda_num]  # 所在组拥有的CUDA设备。

    # * 等待当前CUDA设备空闲。
    if block:
        cuda_block_util_no_process_on(current_cudas, verbosity=verbosity)

    # * 获取当前指定了当前设备的环境变量。
    env_with_current_cuda = os.environ.copy()
    env_with_current_cuda['CUDA_VISIBLE_DEVICES'] = ','.join([str(c) for c in current_cudas])

    # * 返回分配的cuda设备。
    return current_cudas, env_with_current_cuda



# -- configs/mnist/base,sched_from_addon,2xlr,2÷bs/cfg.py --

from alchemy_cat.dl_config import Config

cfg = Config(caps='configs/mnist/base,sched_from_addon/cfg.py')

cfg.loader.ini.batch_size = 256

cfg.sched.epochs = 15



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author  : Xiaobo Yang
@Contact : hal_42@zju.edu.cn
@Time    : 2023/2/4 15:57
@File    : __init__.py.py
@Software: PyCharm
@Desc    : 
"""
from .auto_allocate_cuda import *
from .cuda_block import *



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@author: Xiaobo Yang
@contact: hal_42@zju.edu.cn
@software: PyCharm
@file: __init__.py.py
@time: 2020/1/7 22:15
@desc:
"""
# NOTE figure_wall的导入将触发matplotlib导入，其非常耗时，所以不在这里导入。
# from .figure_wall import *
from .utils import *
from .plot_conf_matrix import pretty_plot_confusion_matrix
from .subplots_row_col import *
from .shuffle_ch import *



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author  : Xiaobo Yang
@Contact : hal_42@zju.edu.cn
@Time    : 2023/3/30 20:18
@File    : mask.py
@Software: PyCharm
@Desc    : 
"""
import torch
import numpy as np

__all__ = ['thresh_mask']


def thresh_mask(data: dict, thresh: float, ignore_label: int=255) -> np.ndarray:
    """根据bg_fg_score的最大值，将小于阈值的seed置为ignore_label。

    Args:
        data: data['seed']为种子，data['bg_fg_score']为前背景得分。
        thresh: 小于阈值的种子点被置为ignore_label。
        ignore_label: 默认255。

    Returns:
        伪真值。
    """
    seed, fg_bg_score = data['seed'].copy(), data['bg_fg_score']  # (H, W), (P+1, H, W)

    max_score = torch.asarray(fg_bg_score, device='cuda').amax(dim=0)  # (H, W)
    ignore_mask = (max_score < thresh).cpu().numpy()

    seed[ignore_mask] = ignore_label

    return seed



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@author: Xiaobo Yang
@contact: hal_42@zju.edu.cn
@software: PyCharm
@file: online_seed.py
@time: 2020/3/29 23:19
@desc:
"""
import numpy as np
import torch

from alchemy_cat.alg.normalize_tensor import channel_min_max_norm
from alchemy_cat.contrib.tasks.wsss.seeding.e2e_seeding.cam_sal_to_seed import cam_sal_to_seed

__all__ = ['get_online_seed_from_cam_sal_seg']


def get_online_seed_from_cam_sal_seg(cam: torch.Tensor, sal: torch.Tensor,
                                     seg_prob: torch.Tensor, cls_in_label: torch.Tensor,
                                     cam_thresh: float, sal_thresh: float, ignore_label: int) -> torch.Tensor:
    """Get online seed from CAM, saliency and segmentation prob output

    Args:
        cam: (N, num_class - 1, H, W) Class Activation Map
        sal: (N, H, W) Saliency
        seg_prob: (N, num_class, H, W) prob output of segmentation branch
        cls_in_label: (N, num_class) one hot label
        cam_thresh: cam lower bound
        sal_thresh: cam upper bound
        ignore_label: label ignored

    Returns:
        Online seeds
    """
    clamp_cam = torch.clamp(cam, min=0.0)

    norm_cam = channel_min_max_norm(clamp_cam)

    comb_cam = norm_cam + seg_prob[:, 1:, ...]

    seeds = []
    for ele_comb_cam, ele_sal, ele_cls_in_label in zip(comb_cam, sal, cls_in_label):
        np_comb_cam, np_sal, np_cls_in_label = ele_comb_cam.cpu().numpy().transpose(1, 2, 0), \
                                               ele_sal.cpu().numpy(), \
                                               ele_cls_in_label.cpu().numpy()

        loc_cue = cam_sal_to_seed(np_comb_cam, np_sal, np_cls_in_label, cam_thresh, sal_thresh, ignore_label)

        seed = np.zeros(seg_prob.shape[1:], dtype=np.int64)
        non_ignore_indices = np.nonzero(loc_cue != ignore_label)
        seed[loc_cue[non_ignore_indices], non_ignore_indices[0], non_ignore_indices[1]] = 1

        seeds.append(seed)

    np_seeds = np.stack(seeds, axis=0)
    return torch.from_numpy(np_seeds).long().cuda(seg_prob.device)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@author: Xiaobo Yang
@contact: hal_42@zju.edu.cn
@software: PyCharm
@file: get_process_info.py
@time: 2020/6/3 0:00
@desc:
"""
from typing import Optional
import psutil

__all__ = ['get_process_info']


def get_process_info(pid: Optional[int]=None, verbosity=0) -> dict:
    """Return a dict with PID's process info.

    Args:
        pid: The PID of process. If None, use current process's PID. (Default: None)
        verbosity: Info verbosity 0-3. (Default: 0)

    Returns:
        Dict with process info
    """
    process_info = {}

    p = psutil.Process(pid)
    all_info = p.as_dict()

    process_info['exe'] = all_info['exe']
    process_info['cmdline'] = all_info['cmdline']
    process_info['cwd'] = all_info['cwd']
    if verbosity > 0:
        process_info['name'] = all_info['name']
        process_info['pid'] = all_info['pid']
        process_info['ppid'] = all_info['ppid']
        process_info['username'] = all_info['username']
        process_info['environ'] = all_info['environ']
    if verbosity > 1:
        process_info['cpu_percent'] = all_info['cpu_percent']
        process_info['nice'] = all_info['nice']
        process_info['memory_info'] = all_info['memory_info']
        process_info['open_files'] = all_info['open_files']
    if verbosity > 2:
        process_info = all_info

    return process_info



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author  : Xiaobo Yang
@Contact : hal_42@zju.edu.cn
@Time    : 2022/5/31 17:32
@File    : rng_cacher.py
@Software: PyCharm
@Desc    : 
"""
from typing import Optional

import os
import os.path as osp

import torch
import torch.distributed as dist

__all__ = ['RNGCacher']


# TODO 增加numpy支持。
class RNGCacher(object):
    """缓存随机状态。"""

    def __init__(self):
        self._rng_state: Optional[torch.Tensor] = None
        self._cuda_rng_state: Optional[torch.Tensor] = None

    def cache(self):
        self._rng_state = torch.get_rng_state()
        self._cuda_rng_state = torch.cuda.get_rng_state()

    def resume(self, purge: bool=True):
        assert self._rng_state is not None
        assert self._cuda_rng_state is not None

        torch.set_rng_state(self._rng_state)
        torch.cuda.set_rng_state(self._cuda_rng_state)

        if purge:
            self._rng_state = None
            self._cuda_rng_state = None

    def __enter__(self):
        """退出后回到进入前的随机状态。"""
        self.cache()

    def __exit__(self, *args):
        self.resume(purge=True)

    def __iter__(self):
        """每次循环开始时，回到循环前随机状态。"""
        self.cache()
        while True:
            yield
            self.resume(purge=False)

    @staticmethod
    def rng_state_file(cache_dir):
        if dist.is_initialized():
            rng_state_file = osp.join(cache_dir, '.rng_state_log', f'rank{dist.get_rank()}_rng_state.pth')
        else:
            rng_state_file = osp.join(cache_dir, '.rng_state_log', 'rng_state.pth')
        return rng_state_file

    @classmethod
    def cache_to_file(cls, cache_dir: str):
        """将当前（rank）随机状态保存到文件。"""
        os.makedirs(osp.join(cache_dir, '.rng_state_log'), exist_ok=True)
        torch.save((torch.get_rng_state(), torch.cuda.get_rng_state()), cls.rng_state_file(cache_dir))

    @classmethod
    def resume_from_file(cls, cache_dir: str):
        """从文件中恢复（当前rank）的随机状态。"""
        rng_state, cuda_rng_state = torch.load(cls.rng_state_file(cache_dir))
        torch.set_rng_state(rng_state)
        torch.cuda.set_rng_state(cuda_rng_state)

    @classmethod
    def resume_except_cache_to_file(cls, cache_dir):
        """尝试从文件中恢复（当前rank）的随机状态，失败则将当前（rank）随机状态保存到文件。"""
        try:
            cls.resume_from_file(cache_dir)
        except Exception:
            print(f"无法从{cls.rng_state_file(cache_dir)}恢复随机状态, 转而保存当前随机状态。")
            cls.cache_to_file(cache_dir)
        else:
            print(f"从{cls.rng_state_file(cache_dir)}恢复随机状态。")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author  : Xiaobo Yang
@Contact : hal_42@zju.edu.cn
@Time    : 2024/2/21 17:06
@File    : loguru_exts.py
@Software: PyCharm
@Desc    : 
"""
import sys

from loguru import logger

__all__ = ['init_loguru']


def init_loguru(**kwargs) -> type(logger):
    """删除此前所有loguru handlers并初始化一个新的、以stdout为输出的loguru logger。

    Args:
        **kwargs: 传递给loguru.add的参数。

    Returns:
        Logger: loguru logger。
    """
    # -* 配置loguru logger。
    config = {"handlers": [{"sink": sys.stdout,
                            "format": '<green>{time:YYYY-MM-DD HH:mm:ss.SSS}</green> | '
                                      '<level>{level: <8}</level> | '
                                      '<cyan>{name}</cyan>:<cyan>{function}</cyan>:<cyan>{line}</cyan> - '
                                      '<level>{message}</level>',
                            "colorize": True,
                            **kwargs}]}
    logger.configure(**config)

    # -* warning以下message不加粗。
    logger.level('TRACE', color='<cyan>')
    logger.level("DEBUG", color='<blue>')
    logger.level("INFO", color='')
    logger.level('SUCCESS', color='<green>')

    return logger



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@author: Xiaobo Yang
@contact: hal_42@zju.edu.cn
@software: PyCharm
@file: boundary_loss.py
@time: 2020/3/28 22:53
@desc:
"""
from typing import Tuple, Union

import numpy as np
import scipy.ndimage as nd
import torch
import torch.nn.functional as F
from torch.multiprocessing import Pool

from alchemy_cat.acplot.shuffle_ch import BGR2RGB

from alchemy_cat.alg.dense_crf import DenseCRF

__all__ = ['Boundaryloss']


class Boundaryloss(object):
    def __init__(self, dense_crf: DenseCRF, eps: float=1e-5, crf_num_workers: int=4):
        """Compute boundary loss

        Args:
            dense_crf: DenseCRF functor
            eps: Min prob allowed when clamp probs
            crf_num_workers: num of workers when crf parallel
        """
        self.dense_crf = dense_crf
        self.eps = eps
        self.crf_num_workers = crf_num_workers

        self.crf_pool = Pool(self.crf_num_workers)

    def crf(self, imgs, probs):
        np_imgs = BGR2RGB(imgs.cpu().numpy().astype(np.uint8).transpose(0, 2, 3, 1))  # (N, H, W, C)
        np_probs = probs.detach().cpu().numpy()  # (N, C, H, W)

        # Scaled imgs to probs shape
        scaled_imgs = nd.zoom(np_imgs,
                              (1.0, np_probs.shape[2] / np_imgs.shape[1], np_probs.shape[3] / np_imgs.shape[2], 1.0),
                              order=1)

        # CRF
        crf_probs = self.crf_pool.starmap(self.dense_crf, zip(scaled_imgs, np_probs))
        crf_prob = np.stack(crf_probs, axis=0)

        # Clamp smoothed probs
        # TODO: Can be removed?
        crf_prob[crf_prob < self.eps] = self.eps
        crf_prob = crf_prob / np.sum(crf_prob, axis=1, keepdims=True)

        # to Tensor
        return torch.from_numpy(crf_prob).float().cuda(probs.get_device())

    def clamp_softmax(self, score, dim=1):
        probs = torch.clamp(F.softmax(score, dim), self.eps, 1)
        probs = probs / torch.sum(probs, dim=dim, keepdim=True)
        return probs

    def __call__(self, images, score_map, out_prob=False) -> Union[torch.Tensor, Tuple[torch.Tensor, torch.Tensor]]:
        """Compute the constrain-to-boundary loss

        Args:
            images: (N, 3, H, W) RGB img
            score_map: (N, C, H, W) score map
            out_prob: If true, return smoothed predict_probs. (Default: False)

        Returns:
            constrain-to-boundary loss
        """
        probs = self.clamp_softmax(score_map)
        smooth_probs = self.crf(images, probs)
        # Compute KL-Div
        # TODO: clamp is not needed?
        loss = torch.mean(torch.sum(smooth_probs * torch.log(torch.clamp(smooth_probs / probs, 0.05, 20)), dim=1))

        if out_prob:
            return loss, smooth_probs

        return loss

    def __del__(self):
        self.crf_pool.close()
        self.crf_pool.join()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author  : Xiaobo Yang
@Contact : hal_42@zju.edu.cn
@Time    : 2023/11/26 22:22
@File    : __init__.py.py
@Software: PyCharm
@Desc    : 
"""
from .coco14seg import *



# -*- coding: utf-8 -*-
import matplotlib.font_manager as fm
import matplotlib.pyplot as plt
import numpy as np
import seaborn as sn
from matplotlib.collections import QuadMesh

__all__ = ['pretty_plot_confusion_matrix']


def get_new_fig(fn, figsize=(9, 9)):
    """Init graphics"""
    fig1 = plt.figure(fn, figsize)
    ax1 = fig1.gca()  # Get Current Axis
    ax1.cla()  # clear existing plot
    return fig1, ax1


def config_cell_text_and_colors(
        array_df, lin, col, oText, facecolors, pos, fz, show_null_values=0
):
    """
    config cell text and colors
    and return text elements to add and to dell
    @TODO: use fmt
    """
    text_add = []
    text_del = []
    cell_val = array_df[lin][col]
    tot_all = array_df[-1][-1]
    per = (float(cell_val) / tot_all) * 100
    curr_column = array_df[:, col]
    ccl = len(curr_column)

    # last line  and/or last column
    if (col == (ccl - 1)) or (lin == (ccl - 1)):
        # tots and percents
        if cell_val != 0:
            if (col == ccl - 1) and (lin == ccl - 1):
                tot_rig = 0
                for i in range(array_df.shape[0] - 1):
                    tot_rig += array_df[i][i]
                per_ok = (float(tot_rig) / cell_val) * 100
            elif col == ccl - 1:
                tot_rig = array_df[lin][lin]
                per_ok = (float(tot_rig) / cell_val) * 100
            elif lin == ccl - 1:
                tot_rig = array_df[col][col]
                per_ok = (float(tot_rig) / cell_val) * 100
            per_err = 100 - per_ok
        else:
            per_ok = per_err = 0

        per_ok_s = ["%.2f%%" % per_ok, "100%"][per_ok == 100]

        # text to DEL
        text_del.append(oText)

        # text to ADD
        font_prop = fm.FontProperties(weight="bold", size=fz)
        text_kwargs = dict(
            color="w",
            ha="center",
            va="center",
            gid="sum",
            fontproperties=font_prop,
        )
        lis_txt = ["%d" % cell_val, per_ok_s, "%.2f%%" % per_err]
        lis_kwa = [text_kwargs]
        dic = text_kwargs.copy()
        dic["color"] = "g"
        lis_kwa.append(dic)
        dic = text_kwargs.copy()
        dic["color"] = "r"
        lis_kwa.append(dic)
        lis_pos = [
            (oText._x, oText._y - 0.3),
            (oText._x, oText._y),
            (oText._x, oText._y + 0.3),
        ]
        for i in range(len(lis_txt)):
            newText = dict(
                x=lis_pos[i][0],
                y=lis_pos[i][1],
                text=lis_txt[i],
                kw=lis_kwa[i],
            )
            text_add.append(newText)

        # set background color for sum cells (last line and last column)
        carr = [0.27, 0.30, 0.27, 1.0]
        if (col == ccl - 1) and (lin == ccl - 1):
            carr = [0.17, 0.20, 0.17, 1.0]
        facecolors[pos] = carr

    else:
        if per > 0:
            txt = "%s\n%.2f%%" % (cell_val, per)
        else:
            if show_null_values == 0:
                txt = ""
            elif show_null_values == 1:
                txt = "0"
            else:
                txt = "0\n0.0%"
        oText.set_text(txt)

        # main diagonal
        if col == lin:
            # set color of the text in the diagonal to white
            oText.set_color("w")
            # set background color in the diagonal to blue
            facecolors[pos] = [0.35, 0.8, 0.55, 1.0]
        else:
            oText.set_color("r")

    return text_add, text_del


def insert_totals(df_cm):
    """insert total column and line (the last ones)"""
    sum_col = []
    for c in df_cm.columns:
        sum_col.append(df_cm[c].sum())
    sum_lin = []
    for item_line in df_cm.iterrows():
        sum_lin.append(item_line[1].sum())
    df_cm["sum_lin"] = sum_lin
    sum_col.append(np.sum(sum_lin))
    df_cm.loc["sum_col"] = sum_col


def pp_matrix(
        df_cm,
        annot=True,
        cmap="Oranges",
        fmt=".2f",
        fz=11,
        lw=0.5,
        cbar=False,
        figsize=(8, 8),
        show_null_values=0,
        pred_val_axis="y",
        norm=None
):
    """
    print conf matrix with default layout (like matlab)
    params:
      df_cm          dataframe (pandas) without totals
      annot          print text in each cell
      cmap           Oranges,Oranges_r,YlGnBu,Blues,RdBu, ... see:
      fz             fontsize
      lw             line width
      pred_val_axis  where to show the prediction values (x or y axis)
                      'col' or 'x': show predicted values in columns (x-axis) instead lines
                      'lin' or 'y': show predicted values in lines   (y-axis)
    """
    if pred_val_axis in ("col", "x"):
        xlbl = "Predicted"
        ylbl = "Actual"
    else:
        xlbl = "Actual"
        ylbl = "Predicted"
        df_cm = df_cm.T

    # create "Total" column
    insert_totals(df_cm)

    # this is for print always in the same window
    fig, ax1 = get_new_fig("Conf matrix default", figsize)

    ax = sn.heatmap(
        df_cm,
        annot=annot,
        annot_kws={"size": fz},
        linewidths=lw,
        ax=ax1,
        cbar=cbar,
        cmap=cmap,
        linecolor="w",
        fmt=fmt,
        norm=norm
    )

    # set tick labels rotation
    ax.set_xticklabels(ax.get_xticklabels(), rotation=45, fontsize=10)
    ax.set_yticklabels(ax.get_yticklabels(), rotation=25, fontsize=10)

    # Turn off all the ticks
    for t in ax.xaxis.get_major_ticks():
        t.tick1On = False
        t.tick2On = False
    for t in ax.yaxis.get_major_ticks():
        t.tick1On = False
        t.tick2On = False

    # face colors list
    quad_mesh = ax.findobj(QuadMesh)[0]
    facecolors = quad_mesh.get_facecolors()

    # iter in text elements
    array_df = np.array(df_cm.to_records(index=False).tolist())
    text_add = []
    text_del = []
    position = -1  # from left to right, bottom to top.
    for t in ax.collections[0].axes.texts:  # ax.texts:
        pos = np.array(t.get_position()) - [0.5, 0.5]
        lin = int(pos[1])
        col = int(pos[0])
        position += 1

        # set text
        txt_res = config_cell_text_and_colors(
            array_df, lin, col, t, facecolors, position, fz, show_null_values
        )

        text_add.extend(txt_res[0])
        text_del.extend(txt_res[1])

    # remove the old ones
    for item in text_del:
        item.remove()
    # append the new ones
    for item in text_add:
        ax.text(item["x"], item["y"], item["text"], **item["kw"])

    # titles and legends
    ax.set_title("Confusion matrix")
    ax.set_xlabel(xlbl)
    ax.set_ylabel(ylbl)
    fig.tight_layout()  # set layout slim
    return fig


pretty_plot_confusion_matrix = pp_matrix



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author  : Xiaobo Yang
@Contact : hal_42@zju.edu.cn
@Time    : 2021/7/8 17:15
@File    : subplots_row_col.py
@Software: PyCharm
@Desc    : 
"""
from typing import Tuple
import math

__all__ = ["square", "row_all", "col_all", "rect"]


def square(num: int) -> Tuple[int, int]:
    col_num = math.ceil(math.sqrt(num))
    row_num = math.ceil(num / col_num)
    return (row_num, col_num)


def row_all(num: int) -> Tuple[int, int]:
    return (num, 1)


def col_all(num: int) -> Tuple[int, int]:
    return (1, num)


def rect(num: int, row_num: int=1) -> Tuple[int, int]:
    col_num = math.ceil(num / row_num)
    return (row_num, col_num)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author  : Xiaobo Yang
@Contact : hal_42@zju.edu.cn
@Time    : 2022/9/7 15:05
@File    : avg_models.py
@Software: PyCharm
@Desc    : 
"""
from tqdm import tqdm

import torch

__all__ = ['avg_states']


def avg_states(state_files: list[str], device: torch.device= 'cpu', verbosity: bool=True) -> dict[str, torch.Tensor]:
    """多个模型参数取平均。

    Args:
        state_files: 模型文件。
        device: 模型加载到哪个设备上。
        verbosity: 若为True，打印合体进程条。

    Returns:
        平均后的模型参数。
    """
    avg_state: dict[str, torch.Tensor] | None = None
    for state_file in (tqdm(state_files, desc='合体', unit='个', dynamic_ncols=True) if verbosity else state_files):
        # * 载入模型。
        state: dict[str, torch.Tensor] = torch.load(state_file, map_location=device)
        # * 第一个模型，作为基础模型。
        if avg_state is None:
            avg_state = state
        # * 此后模型，检查和平均模型键值、参数相同，加到平均模型上。
        else:
            assert set(avg_state.keys()) == set(state.keys())
            for k in avg_state.keys():
                assert avg_state[k].dtype == state[k].dtype
                assert avg_state[k].shape == state[k].shape
                avg_state[k] += state[k]

    # * 模型参数取平均。
    for k, v in avg_state.items():
        avg_state[k] = v / len(state_files) if v.is_floating_point() else torch.div(v, len(state_files),
                                                                                    rounding_mode='floor')

    return avg_state



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author  : Xiaobo Yang
@Contact : hal_42@zju.edu.cn
@Time    : 2021/5/7 10:12
@File    : test_cfg2tune.py
@Software: PyCharm
@Desc    : 
"""
import os
import os.path as osp
import pickle
import shutil
import sys
from typing import Optional

import pytest
try:
    from addict import Dict
except ImportError:
    Dict = dict

sys.path = ['', 'py_tools/param_tuner/tests'] + sys.path  # noqa: E402

from alchemy_cat.py_tools import Cfg2Tune, open_config, ItemLazy, Config


@pytest.fixture(scope="function")
def cfg_dir():
    cfg_dir = osp.join('Temp', 'param_tuner_test_cache')
    yield cfg_dir
    shutil.rmtree(cfg_dir, ignore_errors=True)


@pytest.fixture(scope='function')
def easy_config():
    return Cfg2Tune.load_cfg2tune(osp.join('configs', 'easy_config', 'cfg.py'))


@pytest.fixture(scope='function')
def imported_outside_config():
    return Cfg2Tune.load_cfg2tune(osp.join('configs', 'imported_outside_config', 'cfg.py'))


@pytest.fixture(scope='function')
def def_inside_config():
    sys.path.insert(0, os.getcwd())
    return Cfg2Tune.load_cfg2tune(osp.join('configs', 'def_inside_config', 'cfg.py'))


@pytest.fixture(scope='function')
def mutual_depend_config():
    return Cfg2Tune.load_cfg2tune(osp.join('configs', 'mutual_depend_config', 'cfg.py'))


@pytest.fixture(scope='function')
def no_legal_val_config():
    return Cfg2Tune.load_cfg2tune(osp.join('configs', 'no_legal_val_config', 'cfg.py'))


@pytest.fixture(scope='function')
def no_param_config():
    return Cfg2Tune.load_cfg2tune(osp.join('configs', 'no_param_config', 'cfg.py'))


@pytest.fixture(scope='function')
def run_time_mutual_depend_config():
    return Cfg2Tune.load_cfg2tune(osp.join('configs', 'run_time_mutual_depend_config', 'cfg.py'))


@pytest.fixture(scope='function')
def subject_to_param_config():
    return Cfg2Tune.load_cfg2tune(osp.join('configs', 'subject_to_param_config', 'cfg.py'))


@pytest.fixture(scope='function')
def subject_to_static_config():
    return Cfg2Tune.load_cfg2tune(osp.join('configs', 'subject_to_static_config', 'cfg.py'))


@pytest.fixture(scope='function')
def wrong_sequence_config():
    return Cfg2Tune.load_cfg2tune(osp.join('configs', 'wrong_sequence_config', 'cfg.py'))


@pytest.fixture(scope='function')
def itemlazy_config():
    return Cfg2Tune.load_cfg2tune(osp.join('configs', 'itemlazy_config', 'cfg.py'))


@pytest.fixture(scope='function')
def init_with_config_config():
    return Cfg2Tune.load_cfg2tune(osp.join('configs', 'init_with_config', 'cfg.py'))


@pytest.fixture(scope='function')
def init_with_cfg2tune_config():
    return Cfg2Tune.load_cfg2tune(osp.join('configs', 'init_with_cfg2tune', 'cfg.py'))


@pytest.fixture(scope='function')
def init_with_update_config():
    return Cfg2Tune.load_cfg2tune(osp.join('configs', 'init_with_update', 'cfg.py'))


def test_root_child_params2tune(easy_config):
    assert len(easy_config.ordered_params2tune) == 4


def test_dump_reload(easy_config, cfg_dir):
    cfgs = [cfg.to_dict() for cfg in easy_config.get_cfgs()]
    cfg_pkls, _ = easy_config.dump_cfgs(cfg_dir)

    load_cfgs = []
    for cfg_pkl in cfg_pkls:
        with open(cfg_pkl, "rb") as pkl_f:
            load_cfgs.append(pickle.load(pkl_f))

    assert cfgs == load_cfgs


def test_dump_reload_cfg2tune_with_function_imported_outside(imported_outside_config, cfg_dir):
    """测试生成的cfg_tuned中含有函数的情况。函数从外部库导入，或者在本项目中定义。"""
    cfgs = [cfg.to_dict() for cfg in imported_outside_config.get_cfgs()]
    cfg_pkls, _ = imported_outside_config.dump_cfgs(cfg_dir)

    load_cfgs = []
    for cfg_pkl in cfg_pkls:
        with open(cfg_pkl, "rb") as pkl_f:
            load_cfgs.append(pickle.load(pkl_f))

    assert cfgs == load_cfgs


def test_dump_reload_with_function_define_inside(def_inside_config, cfg_dir):
    """测试生成的cfg_tuned中含有函数的情况。函数在Cfg2Tune所在py文件内定义。"""
    cfgs = [cfg.to_dict() for cfg in def_inside_config.get_cfgs()]
    cfg_pkls, _ = def_inside_config.dump_cfgs(cfg_dir)

    load_cfgs = []
    for cfg_pkl in cfg_pkls:
        with open(cfg_pkl, "rb") as pkl_f:
            load_cfgs.append(pickle.load(pkl_f))

    assert cfgs == load_cfgs


def test_easy_config(easy_config):
    def get_att_val(cfg):
        return [cfg.foo0.foo1.a, cfg.foo0.foo1.b, cfg.foo0.fix, cfg.c, cfg.foo0.a_max]

    wanted_att_val = [
        [0, .1, '0', False, 2],
        [0, .1, '0', True, 2],
        [0, .2, '0', False, 2],
        [0, .2, '0', True, 2],
        [1, .1, '0', False, 2],
        [1, .1, '0', True, 2],
        [1, .2, '0', False, 2],
        [1, .2, '0', True, 2],
        [2, .1, '0', False, 2],
        [2, .1, '0', True, 2],
        [2, .2, '0', False, 2],
        [2, .2, '0', True, 2],
    ]

    att_val = [get_att_val(cfg) for cfg in easy_config.get_cfgs()]
    assert wanted_att_val == att_val


def test_easy_config_twice(easy_config):
    def get_att_val(cfg):
        return [cfg.foo0.foo1.a, cfg.foo0.foo1.b, cfg.foo0.fix, cfg.c, cfg.foo0.a_max]

    wanted_att_val = [
        [0, .1, '0', False, 2],
        [0, .1, '0', True, 2],
        [0, .2, '0', False, 2],
        [0, .2, '0', True, 2],
        [1, .1, '0', False, 2],
        [1, .1, '0', True, 2],
        [1, .2, '0', False, 2],
        [1, .2, '0', True, 2],
        [2, .1, '0', False, 2],
        [2, .1, '0', True, 2],
        [2, .2, '0', False, 2],
        [2, .2, '0', True, 2],
    ]

    att_val = [get_att_val(cfg) for cfg in easy_config.get_cfgs()]
    assert wanted_att_val == att_val

    att_val = [get_att_val(cfg) for cfg in easy_config.get_cfgs()]
    assert wanted_att_val == att_val


def test_wrong_sequence_config(wrong_sequence_config):
    with pytest.raises(RuntimeError, match="Check if the former param call later param"):
        next(wrong_sequence_config.get_cfgs())


def test_mutual_depend_config(mutual_depend_config):
    with pytest.raises(RuntimeError, match="Check if the former param call later param"):
        next(mutual_depend_config.get_cfgs())


def test_run_time_mutual_depend_config(run_time_mutual_depend_config):
    cfgs = run_time_mutual_depend_config.get_cfgs()

    for _ in range(2):
        next(cfgs)

    with pytest.raises(RuntimeError, match="Check if the former param call later param"):
        next(cfgs)


def test_no_legal_val_config(no_legal_val_config):
    def get_att_val(cfg):
        return [cfg.foo0.foo1.a, cfg.foo0.foo1.b, cfg.c]

    wanted_att_val = [
        [0, .1, False],
        [0, .1, True],
        [1, .1, False],
        [1, .1, True],
    ]

    att_val = [get_att_val(cfg) for cfg in no_legal_val_config.get_cfgs()]
    assert wanted_att_val == att_val


def test_same_name_config():
    with pytest.raises(RuntimeError, match="for param2tune repeated."):
        Cfg2Tune.load_cfg2tune(osp.join('py_tools/param_tuner/tests', 'configs', 'same_name_config', 'cfg.py'))


def test_no_param_config(no_param_config):
    with pytest.raises(RuntimeError, match="must have more than 1 Param2Tune"):
        next(no_param_config.get_cfgs())


def test_subject_to_static_config(subject_to_static_config):
    def get_att_val(cfg):
        return [cfg.foo0.foo1.a, cfg.foo0.foo1.b, cfg.c]

    wanted_att_val = [
        [0, .1, False],
        [0, .1, True],
        [0, .2, False],
        [0, .2, True],
        [1, .1, False],
        [1, .1, True],
        [1, .2, False],
        [1, .2, True],
    ]

    att_val = [get_att_val(cfg) for cfg in subject_to_static_config.get_cfgs()]
    assert wanted_att_val == att_val


def test_subject_to_param_config(subject_to_param_config):
    def get_att_val(cfg):
        return [cfg.foo0.foo1.a, cfg.foo0.foo1.b, cfg.c]

    wanted_att_val = [
        [0, .1, True],
        [0, .2, True],
        [1, .2, True],
        [2, .1, False],
    ]

    att_val = [get_att_val(cfg) for cfg in subject_to_param_config.get_cfgs()]
    assert wanted_att_val == att_val


def test_itemlazy_config(itemlazy_config, cfg_dir):
    def get_att_val(cfg):
        return [cfg.foo0.foo1.a, cfg.foo0.foo1.b, cfg.foo0.fix, cfg.c, cfg.foo0.a_max, cfg.foo1.lazy]

    wanted_att_val = [
        [0, 0, '0', False, {'a': 1, 'b': 2}, 0],
        [0, 0, '0', True, {'a': 1, 'b': 2}, 0],
        [0, 0, '0', False, {'a': 1, 'b': 2}, 0],
        [0, 1, '0', True, {'a': 1, 'b': 2}, 0],
        [1, 2, '0', False, {'a': 1, 'b': 2}, 2],
        [1, 2, '0', True, {'a': 1, 'b': 2}, 2],
        [1, 0, '0', False, {'a': 1, 'b': 2}, 2],
        [1, 1, '0', True, {'a': 1, 'b': 2}, 2],
        [2, 4, '0', False, {'a': 1, 'b': 2}, 4],
        [2, 4, '0', True, {'a': 1, 'b': 2}, 4],
        [2, 0, '0', False, {'a': 1, 'b': 2}, 4],
        [2, 1, '0', True, {'a': 1, 'b': 2}, 4],
    ]

    cfgs = [open_config(cfg_pkl)[0] for cfg_pkl in itemlazy_config.dump_cfgs(cfg_dir)[0]]

    att_val = [get_att_val(ItemLazy.compute_item_lazy(cfg)) for cfg in cfgs]
    assert wanted_att_val == att_val


def is_self_consistent(cfg: Config, p: Optional[Config]=None, n: Optional[str]=None):
    parent = object.__getattribute__(cfg, '__parent')
    key = object.__getattribute__(cfg, '__key')

    if p is None:
        assert parent is None
        assert key is None
    else:
        assert parent is p
        assert n == key

    for k, v in cfg.items():
        if cfg.is_subtree(v, cfg):
            is_self_consistent(v, cfg, k)


def test_init_with_config(init_with_config_config):
    """检查从Config初始化：能正确转枝赋叶，并找到root。检查配置树自洽性。"""
    leaves = [1, {'a': 2, 'b': 3}, 4, Dict({'c': 5, 'd': 6}), 7, Dict({'e': 8, 'f': 9}), 'init_with_config']

    assert len(list(init_with_config_config.branches)) == 4
    for b in init_with_config_config.branches:
        assert type(b) is Cfg2Tune

    assert len(list(init_with_config_config.leaves)) == len(leaves)
    for l1, l2 in zip(init_with_config_config.leaves, leaves):
        assert l1 == l2
        assert type(l1) is type(l2)

    is_self_consistent(init_with_config_config)


def test_init_with_cfg2tune(init_with_cfg2tune_config):
    """检查从Config初始化：能正确拷枝赋叶，并找到root。检查配置树自洽性。"""
    leaves = [1, Config({'a': 2, 'b': 3}), 4, Dict({'c': 5, 'd': 6}), 7, {'e': 8, 'f': 9}, 'init_with_cfg2tune']

    assert len(list(init_with_cfg2tune_config.branches)) == 4
    for b in init_with_cfg2tune_config.branches:
        assert type(b) is Cfg2Tune

    assert len(list(init_with_cfg2tune_config.leaves)) == len(leaves)
    for l1, l2 in zip(init_with_cfg2tune_config.leaves, leaves):
        assert l1 == l2
        assert type(l1) is type(l2)

    is_self_consistent(init_with_cfg2tune_config)


def test_init_with_update(init_with_update_config):
    """检查从Config初始化：能正确拷枝赋叶，正确更新，并找到root。检查配置树自洽性。"""
    cfg_wanted = Cfg2Tune()

    cfg_wanted.foo0.foo1.a = 1
    cfg_wanted.foo0.foo1.b = {'aa': 2, 'bb': 3}
    cfg_wanted.foo0.foo3.l = 'hh'
    cfg_wanted.foo0.i = 6.8

    cfg_wanted.foo2 = ('c', 'd')

    cfg_wanted.c = 7
    cfg_wanted.d.ee = 8
    cfg_wanted.d.ff = 9
    cfg_wanted.cc = 7
    cfg_wanted.e.f = {10, 'gg'}
    cfg_wanted.e.g = 11

    cfg_wanted.rslt_dir = 'init_with_update'

    # * 从dict角度，检查内容相同。
    assert init_with_update_config == cfg_wanted

    # * 检查枝干和结构相同。
    assert len(list(init_with_update_config.branches)) == len(list(cfg_wanted.branches))
    for b1, b2 in zip(init_with_update_config.branches, cfg_wanted.branches):
        assert b1 == b2
        assert type(b1) is type(b2)

    # * 检查叶子和结构相同。
    assert len(list(init_with_update_config.leaves)) == len(list(cfg_wanted.leaves))
    for l1, l2 in zip(init_with_update_config.leaves, cfg_wanted.leaves):
        assert l1 == l2
        assert type(l1) is type(l2)

    is_self_consistent(init_with_update_config)


def test_init_with_param_lazy():
    """检查从带有ParamLazy的Cfg2Tune初始化。"""
    with pytest.raises(RuntimeError, match="Cfg2Tune can't init by leaf with type"):
        Cfg2Tune('configs/easy_config/cfg.py')


def test_freeze(easy_config):
    """检查冻结与解冻功能。"""
    easy_config.freeze()

    with pytest.raises(RuntimeError, match="is frozen"):
        easy_config.c = 3

    with pytest.raises(RuntimeError, match="is frozen"):
        easy_config.foo0.a_max = 3

    with pytest.raises(RuntimeError, match="is frozen"):
        easy_config.foo2.a = 3


def test_cfg_tuned_subtree(easy_config):
    """检查配置树类型能否正确转换。"""
    cfgs = easy_config.get_cfgs()

    for cfg in cfgs:
        assert type(cfg) is Config
        assert len(list(cfg.branches)) == 3
        assert len(list(cfg.leaves)) == 6



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author  : Xiaobo Yang
@Contact : hal_42@zju.edu.cn
@Time    : 2021/5/21 22:58
@File    : seg_score.py
@Software: PyCharm
@Desc    : 
"""
from typing import Tuple

import numpy as np

from .utils import gather_segment, cam_clip_norm_score

__all__ = ["get_seg_score_by_cam_mean"]


def get_seg_score_by_cam_mean(cam: np.ndarray, segment: np.ndarray, cls_in_label: np.ndarray) \
        -> Tuple[np.ndarray, np.ndarray]:
    fore_in_label = cls_in_label[1:]
    segment_cam, segment_area = gather_segment(cam[fore_in_label.astype(np.bool_)], segment)
    return cam_clip_norm_score(segment_cam), segment_area



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author  : Xiaobo Yang
@Contact : hal_42@zju.edu.cn
@Time    : 2021/6/16 22:49
@File    : pixel_score.py
@Software: PyCharm
@Desc    : 
"""
import numpy as np

from .utils import cam_clip_norm_score, get_fg_cam

__all__ = ["get_pixel_score_by_cam"]


def get_pixel_score_by_cam(cam: np.ndarray, cls_in_label: np.ndarray) -> np.ndarray:
    fg_cam = get_fg_cam(cam, cls_in_label)
    return cam_clip_norm_score(fg_cam)



from .utils import IntermediateLayerGetter
from .deeplab import DeepLabHead, DeepLabHeadV3Plus, DeepLabV3ResNet
from .backbone import resnet, mobilenetv2


def _segm_resnet(name, backbone_name, num_classes, output_stride, pretrained_backbone):
    if output_stride == 4:
        replace_stride_with_dilation = [True, True, True]
        aspp_dilate = [24, 48, 72]
    elif output_stride == 8:
        replace_stride_with_dilation = [False, True, True]
        aspp_dilate = [12, 24, 36]
    elif output_stride == 16:
        replace_stride_with_dilation = [False, False, True]
        aspp_dilate = [6, 12, 18]
    else:
        raise ValueError(f"output_stride = {output_stride} must be 4, 8 or 16.")

    backbone = resnet.__dict__[backbone_name](
        pretrained=pretrained_backbone,
        replace_stride_with_dilation=replace_stride_with_dilation)

    inplanes = 2048
    low_level_planes = 256

    if name == 'deeplabv3plus':
        return_layers = {'layer4': 'out', 'layer1': 'low_level'}
        head = DeepLabHeadV3Plus(inplanes, low_level_planes, num_classes, aspp_dilate)
    elif name == 'deeplabv3':
        return_layers = {'layer4': 'out'}
        head = DeepLabHead(inplanes, num_classes, aspp_dilate)
    else:
        raise ValueError(f"name = {name} must be 'deeplabv3plus' or 'deeplabv3'.")
    backbone = IntermediateLayerGetter(backbone, return_layers=return_layers)

    model = DeepLabV3ResNet(backbone, head)
    return model


def _segm_mobilenet(name, backbone_name, num_classes, output_stride, pretrained_backbone):
    if output_stride == 8:
        aspp_dilate = [12, 24, 36]
    else:
        aspp_dilate = [6, 12, 18]

    backbone = mobilenetv2.mobilenet_v2(pretrained=pretrained_backbone, output_stride=output_stride)

    # rename layers
    backbone.low_level_features = backbone.features[0:4]
    backbone.high_level_features = backbone.features[4:-1]
    backbone.features = None
    backbone.classifier = None

    inplanes = 320
    low_level_planes = 24

    if name == 'deeplabv3plus':
        return_layers = {'high_level_features': 'out', 'low_level_features': 'low_level'}
        classifier = DeepLabHeadV3Plus(inplanes, low_level_planes, num_classes, aspp_dilate)
    elif name == 'deeplabv3':
        return_layers = {'high_level_features': 'out'}
        classifier = DeepLabHead(inplanes, num_classes, aspp_dilate)
    else:
        raise ValueError(f"name = {name} must be 'deeplabv3plus' or 'deeplabv3'.")
    backbone = IntermediateLayerGetter(backbone, return_layers=return_layers)

    model = DeepLabV3ResNet(backbone, classifier)
    return model


def _load_model(arch_type, backbone, num_classes, output_stride, pretrained_backbone):
    if backbone == 'mobilenetv2':
        model = _segm_mobilenet(arch_type, backbone, num_classes, output_stride=output_stride,
                                pretrained_backbone=pretrained_backbone)
    elif backbone.startswith('resnet'):
        model = _segm_resnet(arch_type, backbone, num_classes, output_stride=output_stride,
                             pretrained_backbone=pretrained_backbone)
    else:
        raise NotImplementedError
    return model


# Deeplab v3

def deeplabv3_resnet50(num_classes=21, output_stride=8, pretrained_backbone=True):
    """Constructs a DeepLabV3 model with a ResNet-50 backbone.

    Args:
        num_classes (int): number of classes.
        output_stride (int): output stride for deeplab.
        pretrained_backbone (bool): If True, use the pretrained backbone.
    """
    return _load_model('deeplabv3', 'resnet50', num_classes, output_stride=output_stride,
                       pretrained_backbone=pretrained_backbone)


def deeplabv3_resnet101(num_classes=21, output_stride=8, pretrained_backbone=True):
    """Constructs a DeepLabV3 model with a ResNet-101 backbone.

    Args:
        num_classes (int): number of classes.
        output_stride (int): output stride for deeplab.
        pretrained_backbone (bool): If True, use the pretrained backbone.
    """
    return _load_model('deeplabv3', 'resnet101', num_classes, output_stride=output_stride,
                       pretrained_backbone=pretrained_backbone)


def deeplabv3_mobilenet(num_classes=21, output_stride=8, pretrained_backbone=True):
    """Constructs a DeepLabV3 model with a MobileNetv2 backbone.

    Args:
        num_classes (int): number of classes.
        output_stride (int): output stride for deeplab.
        pretrained_backbone (bool): If True, use the pretrained backbone.
    """
    raise NotImplementedError("Currently not support mobile net.")
    # return _load_model('deeplabv3', 'mobilenetv2', num_classes, output_stride=output_stride,
    #                    pretrained_backbone=pretrained_backbone)


# Deeplab v3+

def deeplabv3plus_resnet50(num_classes=21, output_stride=8, pretrained_backbone=True):
    """Constructs a DeepLabV3 model with a ResNet-50 backbone.

    Args:
        num_classes (int): number of classes.
        output_stride (int): output stride for deeplab.
        pretrained_backbone (bool): If True, use the pretrained backbone.
    """
    return _load_model('deeplabv3plus', 'resnet50', num_classes, output_stride=output_stride,
                       pretrained_backbone=pretrained_backbone)


def deeplabv3plus_resnet101(num_classes=21, output_stride=8, pretrained_backbone=True):
    """Constructs a DeepLabV3+ model with a ResNet-101 backbone.

    Args:
        num_classes (int): number of classes.
        output_stride (int): output stride for deeplab.
        pretrained_backbone (bool): If True, use the pretrained backbone.
    """
    return _load_model('deeplabv3plus', 'resnet101', num_classes, output_stride=output_stride,
                       pretrained_backbone=pretrained_backbone)


def deeplabv3plus_mobilenet(num_classes=21, output_stride=8, pretrained_backbone=True):
    """Constructs a DeepLabV3+ model with a MobileNetv2 backbone.

    Args:
        num_classes (int): number of classes.
        output_stride (int): output stride for deeplab.
        pretrained_backbone (bool): If True, use the pretrained backbone.
    """
    raise NotImplementedError("Currently not support mobile net.")
    # return _load_model('deeplabv3plus', 'mobilenetv2', num_classes, output_stride=output_stride,
    #                    pretrained_backbone=pretrained_backbone)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author  : Xiaobo Yang
@Contact : hal_42@zju.edu.cn
@Time    : 2023/3/16 15:37
@File    : score.py
@Software: PyCharm
@Desc    : 
"""
import numpy as np
import torch
import torch.nn.functional as F
from alchemy_cat.alg import size2HW, resize_cam, resize_cam_cuda, min_max_norm

__all__ = ['cam2score', 'cam2score_cuda', 'cat_bg_score', 'cat_bg_score_cuda', 'idx2seed', 'idx2seed_cuda']


def cam2score(cam: np.ndarray, dsize=None, resize_first: bool=True) -> np.ndarray:
    h, w = size2HW(dsize) if dsize is not None else cam.shape[1:]

    if resize_first:
        score = resize_cam(cam, (h, w))
        score = min_max_norm(score, dim=(1, 2), thresh=0.)
    else:
        score = min_max_norm(cam, dim=(1, 2), thresh=0.)
        score = resize_cam(score, (h, w))

    return score


def cam2score_cuda(cam: torch.Tensor, dsize=None, resize_first: bool=True) -> torch.Tensor:
    h, w = size2HW(dsize) if dsize is not None else cam.shape[1:]

    if resize_first:
        score = resize_cam_cuda(cam, (h, w))
        score = min_max_norm(score, dim=(1, 2), thresh=0.)
    else:
        score = min_max_norm(cam, dim=(1, 2), thresh=0.)
        score = resize_cam_cuda(score, (h, w))

    return score


def cat_bg_score(fg_score: np.ndarray, bg_method: dict) -> np.ndarray:
    """计算背景得分并拼接。

    Args:
        fg_score: (C, H, W) 的得分。
        bg_method: 背景得分计算方法。

    Returns:
        (C+1, H, W) 的前背景得分。
    """
    match bg_method:
        case {'method': 'thresh', 'thresh': thresh}:
            bg_score = np.full_like(fg_score[0], thresh)[None, ...]
        case {'method': 'pow', 'pow': p}:
            bg_score = np.power(1 - np.max(fg_score, axis=0, keepdims=True), p)
        case {'method': 'no_bg'}:  # 无需添加bg，为兼容性而设。
            bg_score = np.empty((0, *fg_score.shape[1:]), dtype=fg_score.dtype)
        case {'method': 'alpha_bg', 'alpha': alpha}:  # bg存在，调整bg得分（做alpha变换）。
            fg_score = fg_score.copy()
            fg_score[0, ...] = fg_score[0, ...] ** alpha
            bg_score = np.empty((0, *fg_score.shape[1:]), dtype=fg_score.dtype)
        case _:
            raise ValueError(f'Unknown bg_method: {bg_method}')

    bg_fg_score = np.concatenate((bg_score, fg_score), axis=0)

    return bg_fg_score


def cat_bg_score_cuda(fg_score: torch.Tensor, bg_method: dict) -> torch.Tensor:
    """计算背景得分并拼接。

    Args:
        fg_score: (C, H, W) 的得分。
        bg_method: 背景得分计算方法。

    Returns:
        (C+1, H, W) 的前背景得分。
    """
    match bg_method:
        case {'method': 'thresh', 'thresh': thresh}:
            bg_score = torch.full_like(fg_score[0], thresh)[None, ...]
        case {'method': 'pow', 'pow': p}:
            bg_score = torch.pow(1 - torch.amax(fg_score, dim=0, keepdim=True), p)
        case {'method': 'no_bg'}:  # 无需添加bg，为兼容性而设。
            bg_score = torch.empty((0, *fg_score.shape[1:]), device=fg_score.device, dtype=fg_score.dtype)
        case {'method': 'alpha_bg', 'alpha': alpha}:  # bg存在，调整bg得分（做alpha变换）。
            fg_score = fg_score.clone()
            fg_score[0, ...] = fg_score[0, ...] ** alpha
            bg_score = torch.empty((0, *fg_score.shape[1:]), device=fg_score.device, dtype=fg_score.dtype)
        case _:
            raise ValueError(f'Unknown bg_method: {bg_method}')

    bg_fg_score = torch.cat((bg_score, fg_score), dim=0)

    return bg_fg_score


def idx2seed(idx: np.ndarray, fg_cls: np.ndarray) -> np.ndarray:
    """将索引张量寻址为对应的种子张量。

    Args:
        idx: (H, W) 的索引张量。
        fg_cls: (C,) 的前景类别张量。

    Returns:
        (H, W) 的种子张量。
    """
    cls_lb = np.pad(fg_cls + 1, (1, 0), mode='constant', constant_values=0)
    seed = cls_lb[idx]

    return seed


def idx2seed_cuda(idx: torch.Tensor, fg_cls: torch.Tensor) -> torch.Tensor:
    """将索引张量寻址为对应的种子张量。

    Args:
        idx: (H, W) 的索引张量。
        fg_cls: (C,) 的前景类别张量。

    Returns:
        (H, W) 的种子张量。
    """
    cls_lb = F.pad(fg_cls + 1, (1, 0), mode='constant', value=0)
    seed = cls_lb[idx]

    return seed



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author  : Xiaobo Yang
@Contact : hal_42@zju.edu.cn
@Time    : 2022/5/25 21:38
@File    : cfg.py
@Software: PyCharm
@Desc    : 测试人有树、叶我无、人有树我有叶，人有叶我有树，人有树我有树。
"""
from alchemy_cat.py_tools import Config


cfg = config = Config()

# * 二级子树。
# 人有树我有树。
cfg.foo0.foo1.b = {'aa': 2, 'bb': 3}  # 人有叶我有叶。
cfg.foo0.foo3.l = 'hh'  # 人有树我无。
cfg.foo0.i = 6.8  # 人有叶我无。
# * 一级子树。
# 人有叶我有树。
cfg.foo2 = ('c', 'd')
# * 叶子。
cfg.cc = 7  # 人有叶我无。
cfg.d = Config({'ee': 8, 'ff': 9})  # 人有树我有叶。
cfg.e.f = {10, 'gg'}  # 人有树我无。
cfg.e.g = 11



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author  : Xiaobo Yang
@Contact : hal_42@zju.edu.cn
@Time    : 2021/4/27 15:13
@File    : undersegmentation_error.py
@Software: PyCharm
@Desc    : 
"""
from typing import Optional

import numpy as np

from alchemy_cat.py_tools import Tracker, Statistic

__all__ = ["UndersegmentationError"]


class UndersegmentationError(Tracker):

    def __init__(self, num_class: int, ignore_label: Optional[int]=255, label_thresh: float=0.2):
        """Compute undersegmentation error of bg, fg and total pixels. Compute average superpixels which intersect with
        fg region.

        Args:
            num_class: Number of classes.
            ignore_label: Ignore label in gt.
            label_thresh: If Superpixel's intersection area to gt > superpixel_area * label_thresh, this superpixel is
                considered to labeled with gt. One superpixel can have multiple gt label.
        """
        # ! Attention, ignore label was computed into bg errors.
        super().__init__()

        self.num_class = num_class

        self._cls_errors = [list() for _ in range(num_class - 1)]
        self._cls_fg_errors = [list() for _ in range(num_class - 1)]
        self._cls_bg_errors = [list() for _ in range(num_class - 1)]
        self._cls_sp_nums_inter_gt = [list() for _ in range(num_class - 1)]

        self.ignore_label = ignore_label
        self.label_thresh = label_thresh

    def update(self, superpixel: np.ndarray, gt: np.ndarray):
        super().update(superpixel, gt)

        # * superpixel should start from 0
        assert np.amin(superpixel) == 0
        # * gt and superpixel should have same shape
        assert superpixel.shape == gt.shape

        # * Get cls in label, drop background
        cls_in_label = np.unique(gt)
        if cls_in_label[0] == 0:
            cls_in_label = cls_in_label[1:]
        if self.ignore_label is not None and cls_in_label[-1] == self.ignore_label:
            cls_in_label = cls_in_label[:-1]

        # * Get one-hot gt & gt_area
        one_hot_gt = (cls_in_label[:, None, None] == gt[None, ...]).astype(np.uint8)
        gt_area = np.sum(one_hot_gt, axis=(1, 2))

        # * Get one-hot superpixel & sp_area
        sp_num = np.amax(superpixel) + 1

        one_hot_sp = np.zeros((sp_num, ) + superpixel.shape, dtype=np.uint8)
        pixel_indices = np.nonzero(np.ones_like(superpixel))
        one_hot_sp[superpixel.reshape(-1), pixel_indices[0], pixel_indices[1]] = 1

        sp_area = np.sum(one_hot_sp, axis=(1, 2))

        # * Get sp gt intersection num
        sp_gt_inter_num = np.sum(one_hot_sp[:, None, ...] * one_hot_gt[None, ...], axis=(2, 3))

        # * Get sp fg, bg and total underseg num
        sp_bg_underseg_num = sp_area - np.sum(sp_gt_inter_num, axis=1)
        sp_fg_underseg_num = (sp_area - sp_bg_underseg_num)[:, None] - sp_gt_inter_num
        sp_underseg_num = sp_area[:, None] - sp_gt_inter_num

        # * Compute thresh hold according to gt area & Get sp-gt label mask
        label_thresh_area = sp_area * self.label_thresh
        sp_gt_label_mask = (sp_gt_inter_num > label_thresh_area[:, None]).astype(np.uint8)

        # * Compute gt fg, bg and total underseg num
        gt_bg_underseg_error = np.sum(sp_bg_underseg_num[:, None] * sp_gt_label_mask, axis=0) / gt_area
        gt_fg_underseg_error = np.sum(sp_fg_underseg_num * sp_gt_label_mask, axis=0) / gt_area
        gt_underseg_error = np.sum(sp_underseg_num * sp_gt_label_mask, axis=0) / gt_area

        # * Compute superpixels num who intersect with gt
        sp_num_inter_gt = np.sum(sp_gt_label_mask, axis=0)

        # * Record metrics
        for cls, bg_underseg_error, fg_underseg_error, underseg_error, num_inter in zip(cls_in_label,
                                                                                        gt_bg_underseg_error,
                                                                                        gt_fg_underseg_error,
                                                                                        gt_underseg_error,
                                                                                        sp_num_inter_gt):
            self._cls_bg_errors[cls - 1].append(bg_underseg_error)
            self._cls_fg_errors[cls - 1].append(fg_underseg_error)
            self._cls_errors[cls - 1].append(underseg_error)
            self._cls_sp_nums_inter_gt[cls - 1].append(num_inter)

    @staticmethod
    def _average_intra_cls(lst: list):
        return [np.mean(cls_lst) for cls_lst in lst]

    @Statistic.getter(importance=0)
    def cls_fg_underseg_error(self):
        return self._average_intra_cls(self._cls_fg_errors)

    @Statistic.getter(importance=0)
    def cls_bg_underseg_error(self):
        return self._average_intra_cls(self._cls_bg_errors)

    @Statistic.getter(importance=0)
    def cls_underseg_error(self):
        return self._average_intra_cls(self._cls_errors)

    @Statistic.getter(importance=0)
    def cls_sp_num_inter_gt(self):
        return self._average_intra_cls(self._cls_sp_nums_inter_gt)

    @Statistic.getter(importance=0)
    def fg_underseg_error(self):
        return np.mean(self.cls_fg_underseg_error)

    @Statistic.getter(importance=0)
    def bg_underseg_error(self):
        return np.mean(self.cls_bg_underseg_error)

    @Statistic.getter(importance=0)
    def underseg_error(self):
        return np.mean(self.cls_underseg_error)

    @Statistic.getter(importance=0)
    def sp_num_inter_gt(self):
        return np.mean(self.cls_sp_num_inter_gt)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@author: Xiaobo Yang
@contact: hal_42@zju.edu.cn
@software: PyCharm
@file: __init__.py
@time: 2020/1/16 4:46
@desc:
"""
from .env import *
from .average_value_tracker import *
from .intermediate_value_getter import *
from .network import *
from .val_env import *
from .set_network import *
from .cuda_tools import *
from .swa import *
from .dist_communication import *
from .rng_cacher import *
from .loss_items import *
from .to_cuda import *



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@author: Xiaobo Yang
@contact: hal_42@zju.edu.cn
@software: PyCharm
@file: eval_preds.py
@time: 2020/4/6 0:01
@desc:
"""
from typing import Optional, Iterable, Callable, Union, Tuple
import os
from os import path as osp
import pickle
from collections import OrderedDict
from tqdm import tqdm

import cv2
import numpy as np

from alchemy_cat.contrib.metrics import SegmentationMetric
from alchemy_cat.py_tools import OneOffTracker

__all__ = ['eval_preds']


def eval_preds(class_num: int, class_names: Optional[Iterable[str]],
               preds_dir: str | list[str], gts_dir: str,
               preds_ignore_label: int=255, gts_ignore_label: int=255,
               result_dir: Optional[str]=None,
               pred_preprocess: Callable[[np.ndarray], np.ndarray]=lambda x: x,
               gt_preprocess: Callable[[np.ndarray], np.ndarray]=lambda x: x,
               importance: int=2,
               eval_individually: bool=True, take_pred_ignore_as_a_cls: bool=False,
               metric_cls: type=SegmentationMetric) \
        -> Union[Tuple[SegmentationMetric, OrderedDict], SegmentationMetric]:

    """Evaluate predictions of semantic segmentation

    Args:
        class_num: Num of classes
        class_names: Name of classes
        preds_dir: Dictionary where preds stored, if list, each element is a pred file path.
        gts_dir: Dictionary where ground truths stored
        preds_ignore_label: Ignore label for predictions
        gts_ignore_label: Ignore label for ground truths
        result_dir: If not None, eval result will be saved to result_dir. (Default: None)
        pred_preprocess: Preprocess function for prediction read in
        gt_preprocess: Preprocess function for ground truth read in
        importance: Segmentation Metric's importance filter. (Default: 2)
        eval_individually: If True, evaluate each sample. (Default: True)
        take_pred_ignore_as_a_cls: If True, the ignored label in preds will be seemed as a class. (Default: False)
        metric_cls: Use metric_cls(class_num, class_names) to eval preds. (Default: SegmentationMetric)

    Returns:
        Segmentation Metric and metric result for each sample (If eval_individually is True)
    """
    assert preds_ignore_label >= class_num
    assert gts_ignore_label >= class_num

    if take_pred_ignore_as_a_cls:
        class_num += 1
        class_names += ['pred_ignore']

    metric = metric_cls(class_num, class_names)
    if eval_individually:
        sample_metrics = OrderedDict()

    print("\n================================== Eval ==================================")
    if isinstance(preds_dir, list) or isinstance(preds_dir, tuple):
        pred_files = preds_dir
    else:
        pred_files = [osp.join(preds_dir, file) for file in os.listdir(preds_dir)]
    for pred_file in tqdm(pred_files,
                          total=len(pred_files),
                          desc='eval progress', unit='sample', dynamic_ncols=True):
        # Set files
        pred_file_suffix = osp.basename(pred_file)
        id = pred_file_suffix.split('.')[0]
        gt_file = osp.join(gts_dir, pred_file_suffix)

        # Read files
        pred = cv2.imread(pred_file, cv2.IMREAD_GRAYSCALE)
        assert pred is not None
        gt = cv2.imread(gt_file, cv2.IMREAD_GRAYSCALE)
        assert gt is not None

        assert pred.shape == gt.shape

        # Preprocess
        pred: np.ndarray = pred_preprocess(pred)
        gt: np.ndarray = gt_preprocess(gt)

        if take_pred_ignore_as_a_cls:
            pred[pred == preds_ignore_label] = class_num - 1
            if gts_ignore_label == class_num - 1:
                gt[gt == gts_ignore_label] = class_num

        # Evaluate
        metric.update(pred, gt)
        if eval_individually:
            with OneOffTracker(lambda: metric_cls(class_num, class_names)) as individual_metric:
                individual_metric.update(pred, gt)
            sample_metrics[id] = individual_metric.statistics(importance)

    # Saving
    if result_dir is not None:
        metric.save_metric(result_dir, importance, dpi=400)
        if eval_individually:
            with open(osp.join(result_dir, 'sample_statistics.pkl'), 'wb') as f:
                pickle.dump(sample_metrics, f)

    print("\n================================ Eval End ================================")

    if eval_individually:
        return metric, sample_metrics
    else:
        return metric



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author  : Xiaobo Yang
@Contact : hal_42@zju.edu.cn
@Time    : 2021/4/17 15:22
@File    : one_hot_numeric_seed_convert.py
@Software: PyCharm
@Desc    : 
"""
import numpy as np

__all__ = ['one_hot_seed_to_numeric_seed', 'numeric_seed_to_one_hot_seed']


def one_hot_seed_to_numeric_seed(one_hot_seed: np.ndarray, ignore_label: int=255) -> np.ndarray:
    """Convert one hot seed to numeric seed

    Args:
        one_hot_seed: (N, C, H, W) one hot seed where (n, c, h, w) = 1 indicates sample n's (h, w) pixel should
            be c class.
        ignore_label: numeric value of ignore label. (Default: 255)

    Returns:
        (N, H, W) numeric seed where (n, h, w) = c means sample n's (h, w) pixel is c class.
    """
    assert np.all((np.sum(one_hot_seed, axis=0) <= 1) & np.sum(one_hot_seed, axis=0) >= 0)

    numeric_seed = np.ones((one_hot_seed.shape[0], one_hot_seed.shape[2], one_hot_seed.shape[3]),
                           dtype=np.uint8) * ignore_label

    seed_loc = np.nonzero(one_hot_seed)
    numeric_seed[seed_loc[0], seed_loc[2], seed_loc[3]] = seed_loc[1]

    return numeric_seed


def numeric_seed_to_one_hot_seed(numeric_seed: np.ndarray, cls_num: int, ignore_label: int=255) -> np.ndarray:
    """Convert numeric seed to one-hot seed

    Args:
        numeric_seed: (N, H, W) numeric seed where (n, h, w) = c indicates sample n's (h, w) pixel should be c class.
        cls_num: number of classes.
        ignore_label: numeric value of ignore label. (Default: 255)

    Returns:
        (N, C, H, W) one-hot seed where (n, c, h, w) = 1 indicates sample n's (h, w) pixel should be c class.
    """
    one_hot_seed = np.zeros((numeric_seed.shape[0], cls_num, numeric_seed.shape[1], numeric_seed.shape[2]),
                            dtype=np.uint8)
    non_ignore_indices = np.nonzero(numeric_seed != ignore_label)
    one_hot_seed[non_ignore_indices[0], numeric_seed[non_ignore_indices],
                 non_ignore_indices[1], non_ignore_indices[2]] = 1
    return one_hot_seed



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@author: Xiaobo Yang
@contact: hal_42@zju.edu.cn
@software: PyCharm
@file: __init__.py
@time: 2020/1/8 4:18
@desc:
"""
from .const import Const
from .process_bar import ProcessBar
from .quick_init import quick_init
from .timer import *
from .func_tools import *
from .str_formatters import *
from .random import *
from .tracker import *
from .numpy_json_encoder import *
from .logger import Logger
from .get_process_info import *
from .find_files import *
from .color_print import *
from .decorators import *
from .check_img_exits import *
from .load_module import *
from .cat_head import *
from .parse_log import *
from .file_md5 import *
# from .cache_dir import *  # python 3.9不兼容，暂时不导入。
from .loguru_exts import *
from .config import *
from .param_tuner import *



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author  : Xiaobo Yang
@Contact : hal_42@zju.edu.cn
@Time    : 2021/5/23 19:52
@File    : solve_conflict.py
@Software: PyCharm
@Desc    : 
"""
from typing import Optional

import numpy as np

from .utils import cls_in_label2tag_always_bg

__all__ = ["solve_conflict_by_area", "solve_conflict_on_channel"]


def solve_conflict_by_area(loc_cue: np.ndarray, cls_in_label: np.ndarray, area: Optional[np.ndarray]=None,
                           ignore_label: int=255, train_boat: bool=True) -> np.ndarray:
    """Solve location cue's conflict by make class with small area to have higher priority.

    Args:
        loc_cue: (fore_class + 1, ...) location cue.
        cls_in_label: (class_num, ) image level label.
        area: The same shape with loc_cue, indicate location's area. If None, set all area to 1. (Default: None)
        ignore_label: int ignore label. (Default: 255)
        train_boat: If True, suppress train_boat's priority after background. (Default: True)

    Returns:
        (...) numeric pred.
    """
    pred = np.ones(loc_cue.shape[1:], dtype=np.int64) * ignore_label

    bg_loc_cue, fg_loc_cue = loc_cue[0], loc_cue[1:]
    fg_tag = cls_in_label2tag_always_bg(cls_in_label)[1:]

    # * Generate background pred
    pred[bg_loc_cue == 1] = 0

    # * Generate foreground seed in the order of their area
    # ** Compute fg area
    area = area if area is not None else np.ones(loc_cue.shape[1:], dtype=np.int64)
    fg_area = np.sum((fg_loc_cue * area).reshape(fg_loc_cue.shape[0], -1), axis=-1)
    # ** Plot cue on pred
    argsort_idx = np.argsort(fg_area)[::-1]  # area in descending order
    for idx in argsort_idx:
        if fg_area[idx] == 0:
            break
        pred[fg_loc_cue[idx] == 1] = fg_tag[idx]

    if train_boat:
        train_boat_ind = ((pred == 4) | (pred == 19)) & (bg_loc_cue == 1)
        pred[train_boat_ind] = 0

    return pred


def solve_conflict_on_channel(loc_cue: np.ndarray, cls_in_label: np.ndarray,
                              channelwise_pred: np.ndarray, ignore_label: int=255) -> np.ndarray:
    """Solve location cue's conflict by run channelwise prediction.

    Args:
        loc_cue: (fore_class + 1, ...) location cue.
        cls_in_label: (class_num, ) image level label.
        channelwise_pred: Channelwise prediction.
        ignore_label: int ignore label. (Default: 255)

    Returns:
        (...) numeric pred.
    """
    pred = np.ones(loc_cue.shape[1:], dtype=np.int64) * ignore_label

    loc_cue_sum = np.sum(loc_cue, axis=0)
    confident_loc = loc_cue_sum == 1
    conflict_loc = loc_cue_sum > 1

    tag = cls_in_label2tag_always_bg(cls_in_label)

    pred[confident_loc] = tag[np.argmax(loc_cue, axis=0), ][confident_loc]  # bg will have sum == 0
    pred[conflict_loc] = channelwise_pred[conflict_loc]

    return pred



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author  : Xiaobo Yang
@Contact : hal_42@zju.edu.cn
@Time    : 2022/7/7 14:56
@File    : cuda_block.py
@Software: PyCharm
@Desc    : 阻塞直至指定CUDA设备释放。
"""
import os
from time import sleep
import warnings

from gpustat import new_query

from alchemy_cat.py_tools import yprint, gprint

__all__ = ['cuda_block_util_no_process_on']

kSleepSecs = 30


def cuda_block_util_no_process_on(cudas: list[str | int] | str, sleep_secs: int=kSleepSecs, verbosity: bool=True):
    """阻塞程序，直到指定的cuda设备上没有任何进程。

    Args:
        cudas: cuda设备列表，可以用字符串列表（如['1', '2']）、整数列表（如[0, 3]）或字符串（如'0,1,2'）表示。
        sleep_secs: 若检测到有进程，间隔多少秒再做检测。
        verbosity: 是否打印阻塞信息。

    Returns:
            无。
    """
    if isinstance(cudas, str):
        cudas = [int(cuda) for cuda in cudas.split(',')]
    else:
        cudas = [int(cuda) for cuda in cudas]

    while True:
        gpu_states = new_query()

        for c, gpu_state in ((cuda, gpu_states[cuda]) for cuda in cudas):
            if len(gpu_state.processes) > 0:
                if verbosity:
                    yprint(f"检测到cuda:{c}上有进程{gpu_state.processes}占用，程序阻塞{sleep_secs}s后，再行检测...")
                break
        else:
            break

        sleep(sleep_secs)

    if verbosity:
        gprint(f"检测到所需CUDA设备{cudas}上已经没有进程占用，阻塞结束。")


if __name__ == "__main__":
    if 'CUDA_VISIBLE_DEVICES' not in os.environ:
        warnings.warn("没有检测到环境变量CUDA_VISIBLE_DEVICES，阻塞无效，退出。")
        exit(0)

    cuda_block_util_no_process_on(os.environ['CUDA_VISIBLE_DEVICES'])



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@author: Xiaobo Yang
@contact: hal_42@zju.edu.cn
@software: PyCharm
@file: numpy_json_encoder.py
@time: 2020/6/2 1:13
@desc:
"""
import json
import numpy as np

__all__ = ['NumpyArrayEncoder']


class NumpyArrayEncoder(json.JSONEncoder):
    """Json encoder for numpy array, which will convert numpy array to list first"""
    def default(self, obj):
        if isinstance(obj, np.integer):
            return int(obj)
        elif isinstance(obj, np.floating):
            return float(obj)
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
        else:
            return super(NumpyArrayEncoder, self).default(obj)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@author: Xiaobo Yang
@contact: hal_42@zju.edu.cn
@software: PyCharm
@file: augers.py
@time: 2020/1/13 9:43
@desc:
"""
from typing import Optional, Union, Iterable, Callable, Tuple, Literal
import numpy as np
import cv2
from collections import abc
from scipy import ndimage as nd
from PIL import Image
import torchvision.transforms as T
import torchvision.transforms.functional as TF

from alchemy_cat.data.data_auger import RandMap, MultiMap
from alchemy_cat.py_tools import Compose, Lambda
from alchemy_cat.py_tools.type import is_int, is_intarr, is_floatarr, tolist
from alchemy_cat.alg import size2HW, color2scalar
from alchemy_cat.acplot.shuffle_ch import RGB2BGR, BGR2RGB

__all__ = ['RandMirror', 'MultiMirror', 'RandUpDown', 'MultiUpDown', 'RandColorJitter', 'scale_img_label',
           'RandScale', 'MultiScale', 'pad_img_label', 'int_img2float32_img', 'centralize', 'RandCrop', 'FiveCrop',
           'RandRangeScale', 'identical', 'pack_identical', 'arr2PIL', 'PIL2arr', 'RandomResizeCrop']


class RandMirror(RandMap):
    rand_seeds = [1, -1]

    def forward(self, img: np.ndarray, label: Optional[np.ndarray] = None) -> Union[np.ndarray, tuple]:
        """ Random mirror image and it's label(if exits)

        Args:
            img (np.ndarray): Img(H, W, C) to be mirrored
            label (Optional[np.ndarray]): Label(H, W) to be mirrored

        Returns: Img and label(if exits) random mirrored
        """
        if label is None:
            return img[:, ::self.rand_seed, :]
        else:
            return img[:, ::self.rand_seed, :], label[:, ::self.rand_seed]


class MultiMirror(MultiMap):
    output_num = 2

    def forward(self, img: np.ndarray, label: Optional[np.ndarray] = None) -> Union[np.ndarray, tuple]:
        """Multi output of mirrored img and label(if exits).
        output_index = 0: Don't mirror
        output_index = 1: mirror

        Args:
            img (np.ndarray): Image(H, W, C) to be mirrored
            label (Optional(np.ndarray)): Label(H, W) to be mirrored

        Returns: Img and label(if exits) random mirrored
        """
        mirror = 1 if self.output_index == 0 else -1

        if label is None:
            return img[:, ::mirror, :]
        else:
            return img[:, ::mirror, :], label[:, ::mirror]


class RandUpDown(RandMap):
    rand_seeds = [1, -1]

    def forward(self, img: np.ndarray, label: Optional[np.ndarray] = None) -> Union[np.ndarray, tuple]:
        """ Random Upside Down image and it's label(if exits)

        Args:
            img (np.ndarray): Img(H, W, C) to be upside down
            label (Optional[np.ndarray]): Label(H, W) to be upside down

        Returns: Img and label(if exits) random upside down
        """
        if label is None:
            return img[::self.rand_seed, :, :]
        else:
            return img[::self.rand_seed, :, :], label[::self.rand_seed, :]


class MultiUpDown(MultiMap):
    output_num = 2

    def forward(self, img: np.ndarray, label: Optional[np.ndarray] = None) -> Union[np.ndarray, tuple]:
        """Multi output of upside down img and label(if exits).
        output_index = 0: Don't upside down
        output_index = 1: upside down

        Args:
            img (np.ndarray): Image(H, W, C) to be upside down
            label (Optional(np.ndarray)): Label(H, W) to be upside down

        Returns: Img and label(if exits) random upside down
        """
        up_down = 1 if self.output_index == 0 else -1

        if label is None:
            return img[::up_down, :, :]
        else:
            return img[::up_down, :, :], label[::up_down, :]


class RandColorJitter(RandMap):

    def __init__(self, max_delta_bright: int = 25, range_mul_contract: tuple = (0.6, 1.4),
                 range_mul_saturate: tuple = (0.5, 1.5), max_delta_hue: int = 25,
                 jitter_prob: Union[Iterable[float], float] = 0.5):
        """Random color jitter for image

        Args:
            max_delta_bright (int): max delta of bright
            range_mul_contract (tuple): range of multiplier of contract
            range_mul_saturate (tuple): range of multiplier of saturate
            max_delta_hue (int): max delta of hue
            jitter_prob (Union[list, tuple, float]): probs of each jitter step (jitter bright, jitter contract,
                jitter saturate, jitter hue) implemented
        """
        super(RandColorJitter, self).__init__()

        if max_delta_bright <= 0:
            raise ValueError(f"max_delta_bright={max_delta_bright} should be larger than 0")
        self.max_delta_bright = max_delta_bright

        if range_mul_contract[0] <= 0 or range_mul_contract[0] >= range_mul_contract[1]:
            raise ValueError(
                f"range_mul_contract={range_mul_contract}'s lower bound should larger than 0, lower than upper bound")
        self.range_mul_contract = range_mul_contract

        if range_mul_saturate[0] <= 0 or range_mul_saturate[0] >= range_mul_saturate[1]:
            raise ValueError(
                f"range_mul_saturate={range_mul_saturate}'s lower bound should larger than 0, lower than upper bound")
        self.range_mul_saturate = range_mul_saturate

        if max_delta_hue <= 0:
            raise ValueError(f"max_delta_hue={max_delta_hue} should be larger than 0")
        self.max_delta_hue = max_delta_hue

        jps = [jitter_prob] * 4 if isinstance(jitter_prob, float) else tolist(jitter_prob)
        if len(jps) != 4:
            raise ValueError(f"If the jitter_prob is not float, then "
                             f"you must give each step's probability(Length of jitter_prob {jitter_prob} must be 4).")
        for jp in jps:
            if jp < 0 or jp > 1:
                raise ValueError(f"jitter probs={jps} should >=0 and <=1")

        self.jitter_prob = jps

    def generate_rand_seed(self, *fwd_args, **fwd_kwargs):
        rand_seed = {}

        is_jitter_bright = np.random.random() < self.jitter_prob[0]
        rand_seed['delta_bright'] = \
            np.random.uniform(-1 * self.max_delta_bright, self.max_delta_bright) if is_jitter_bright else None

        is_jitter_contract = np.random.random() < self.jitter_prob[1]
        rand_seed['mul_contract'] = \
            np.random.uniform(*self.range_mul_contract) if is_jitter_contract else None

        is_jitter_saturate = np.random.random() < self.jitter_prob[2]
        rand_seed['mul_saturate'] = \
            np.random.uniform(*self.range_mul_saturate) if is_jitter_saturate else None

        is_jitter_hue = np.random.random() < self.jitter_prob[3]
        rand_seed['delta_hue'] = \
            np.random.uniform(-1 * self.max_delta_hue, self.max_delta_hue) if is_jitter_hue else None

        jitter_order = list(range(len(rand_seed)))
        np.random.shuffle(jitter_order)
        rand_seed['jitter_order'] = jitter_order

        return rand_seed

    def jitter_bright(self, img: np.ndarray):
        delta_bright = self.rand_seed.get('delta_bright')
        if delta_bright is not None:
            return cv2.convertScaleAbs(img, alpha=1, beta=delta_bright)
        return img

    def jitter_contract(self, img: np.ndarray):
        mul_contract = self.rand_seed.get('mul_contract')
        if mul_contract is not None:
            return cv2.convertScaleAbs(img, alpha=mul_contract, beta=0)
        return img

    def jitter_saturate(self, img: np.ndarray):
        mul_saturate = self.rand_seed.get('mul_saturate')
        if mul_saturate is not None:
            img = cv2.cvtColor(img, cv2.COLOR_BGR2HSV)
            img[:, :, 1] = cv2.convertScaleAbs(img[:, :, 1], alpha=mul_saturate, beta=0)
            img = cv2.cvtColor(img, cv2.COLOR_HSV2BGR)
        return img

    def jitter_hue(self, img: np.ndarray):
        delta_hue = self.rand_seed.get('delta_hue')
        if delta_hue is not None:
            img = cv2.cvtColor(img, cv2.COLOR_BGR2HSV_FULL)
            img[:, :, 0] = cv2.convertScaleAbs(img[:, :, 0], alpha=1, beta=delta_hue)
            img = cv2.cvtColor(img, cv2.COLOR_HSV2BGR_FULL)
        return img

    def forward(self, img: np.ndarray) -> np.ndarray:
        """Color jitter img according to rand seed

        Args:
            img (np.ndarray):  img to be jitter

        Returns: img after jittering
        """
        if not is_intarr(img):
            raise ValueError('Image should be int array')

        jitters = [Lambda(self.jitter_bright), Lambda(self.jitter_contract), Lambda(self.jitter_saturate),
                   Lambda(self.jitter_hue)]
        ordered_jitters = Compose([jitters[i] for i in self.rand_seed['jitter_order']])

        return ordered_jitters(img)


def _check_img_size_equal_label_size(img, label):
    if label is not None and img.shape[:2] != label.shape:
        raise ValueError(f"img size {img.shape[:2]} should be equal to label{label.shape} size")


def scale_img_label(scale_factor: float | tuple[float, float] | int | tuple[int, int],
                    img: np.ndarray, label: Optional[np.ndarray] = None,
                    aligner: Union[Callable[[int], int], Iterable[Callable[[int], int]]] = lambda x: x,
                    align_corner: bool=False, PIL_mode: int | None=None
                    ) -> Union[np.ndarray, Tuple[np.ndarray]]:
    """Scale img and label accroding to scaled factor

    Args:
        scale_factor: scaled factor
        img: img to be scaled
        label: If not none, label to be scaled
        aligner: The scaled_size calculated by scale_factor * img_size will be fix by aligner(scaled_size). If
            Iterable, then first and second aligners separately used to align H and W.
        align_corner: If true, use ndimage's zoom to resize img and label, which can align corner. Else use cv2's
            resize to scaled img and label, which will align center. (Default: False)
        PIL_mode: 若不为None，使用指定的PIL插值模式采样图片。

    Returns: Scaled img and label(If exits)
    """
    _check_img_size_equal_label_size(img, label)

    if PIL_mode is not None and align_corner:
        raise ValueError("使用PIL采样时，align_corner必须是False。")

    if isinstance(aligner, abc.Callable):
        aligner_h, aligner_w = aligner, aligner
    elif isinstance(aligner, abc.Iterable):
        aligner_h, aligner_w = aligner
    else:
        raise ValueError(f"pad_aligner {aligner} must be Callable or Iterable[Callable]")

    match scale_factor:
        case float(s):
            assert s > 0, f"{scale_factor=} 应当 > 0."
            scaled_h, scaled_w = aligner_h(round(s * img.shape[0])), aligner_w(round(s * img.shape[1]))
        case (float(s_h), float(s_w)):
            assert s_h > 0 and s_w > 0, f"{scale_factor=} 应当 > 0."
            scaled_h, scaled_w = aligner_h(round(s_h * img.shape[0])), aligner_w(round(s_w * img.shape[1]))
        case (int() | (int(), int())) as size:
            scaled_h, scaled_w = size2HW(size)
            scaled_h, scaled_w = aligner_h(scaled_h), aligner_w(scaled_w)
        case _:
            raise ValueError(f"不支持的{scale_factor=}。")

    if PIL_mode is not None:
        pil = arr2PIL(img, order='RGB')  # 缩放时，RGB通道顺序无所谓。
        scaled_pil = pil.resize((scaled_w, scaled_h), resample=PIL_mode)
        scaled_img = PIL2arr(scaled_pil, order='RGB')
    elif not align_corner:
        scaled_img = cv2.resize(img, (scaled_w, scaled_h), interpolation=cv2.INTER_LINEAR)
    else:
        scaled_img = nd.zoom(img, (scaled_h / img.shape[0], scaled_w / img.shape[1], 1.0), order=1)

    if label is not None:
        if PIL_mode is not None:
            pil = arr2PIL(label)
            scaled_pil = pil.resize((scaled_w, scaled_h), resample=Image.NEAREST)
            scaled_label = PIL2arr(scaled_pil)
        elif not align_corner:
            scaled_label = cv2.resize(label, (scaled_w, scaled_h), interpolation=cv2.INTER_NEAREST)
        else:
            scaled_label = nd.zoom(label, (scaled_h / label.shape[0], scaled_w / label.shape[1]), order=0)
        return scaled_img, scaled_label
    else:
        return scaled_img


class RandScale(RandMap):

    def __init__(self, scale_factors: Iterable[float | tuple[float, float] | int | tuple[int, int]],
                 aligner: Union[Callable[[int], int], Iterable[Callable[[int], int]]] = lambda x: x,
                 align_corner: bool=False, PIL_mode: int | None=None):
        """Auger to rand rescale the input img and corresponding label

        Args:
            scale_factors: scale_factors: scale factors. eg. [0.5, 1, 1.5] means the img (and label) will be
            rand scale with factor 0.5, 1.0 or 1.5
            aligner: The scaled_size calculated by scale_factor * img_size will be fix by aligner(scaled_size). If
                Iterable, then first and second aligners separately used to align H and W.
            align_corner: If true, use ndimage's zoom to resize img and label, which can align corner. Else use cv2's
                resize to scaled img and label, which will align center. (Default: False)
            PIL_mode: 若不为None，使用指定的PIL插值模式采样图片。
        """
        super(RandScale, self).__init__()

        # self.rand_seeds = []
        # for factor in scale_factors:
        #     factor = float(factor)
        #     if factor <= 0:
        #         raise ValueError(f"scale factors {scale_factors} must all larger than 0")
        #     self.rand_seeds.append(factor)
        self.rand_seeds = list(scale_factors)

        self.aligner = aligner
        self.align_corner = align_corner
        self.PIL_mode = PIL_mode

    def forward(self, img: np.ndarray, label: Optional[np.ndarray] = None) -> Union[np.ndarray, Tuple[np.ndarray]]:
        """Rand Scale img according to rand seed

        Args:
            img (np.ndarray): img with shape (H, W, C)
            label (Optional[np.ndarray]): label with shape (H, W)

        Returns: Scaled img and label(if exit)
        """
        scale_factor = self.rand_seed

        return scale_img_label(scale_factor, img, label, self.aligner,
                               align_corner=self.align_corner, PIL_mode=self.PIL_mode)


class MultiScale(MultiMap):
    def __init__(self, scale_factors: Iterable,
                 aligner: Union[Callable[[int], int], Iterable[Callable[[int], int]]] = lambda x: x,
                 align_corner: bool=False, PIL_mode: int | None=None):
        """Auger to Multi rescale the input img and corresponding label

        Args:
            scale_factors: scale_factors: scale factors. eg. [0.5, 1, 1.5] means the img (and label) will be
            rand scale with factor 0.5, 1.0 or 1.5
            aligner: The scaled_size calculated by scale_factor * img_size will be fix by aligner(scaled_size). If
                Iterable, then first and second aligners separately used to align H and W.
            align_corner: If true, use ndimage's zoom to resize img and label, which can align corner. Else use cv2's
                resize to scaled img and label, which will align center. (Default: False)
            PIL_mode: 若不为None，使用指定的PIL插值模式采样图片。
        """
        super(MultiScale, self).__init__()

        self.scale_factors = []
        for factor in scale_factors:
            factor = float(factor)
            if factor <= 0:
                raise ValueError(f"scale factors {scale_factors} must all larger than 0")
            self.scale_factors.append(factor)

        self.output_num = len(self.scale_factors)

        self.aligner = aligner
        self.align_corner = align_corner
        self.PIL_mode = PIL_mode

    def forward(self, img: np.ndarray, label: Optional[np.ndarray] = None) -> Union[np.ndarray, Tuple[np.ndarray]]:
        """Rand Scale img according to rand seed

        Args:
            img (np.ndarray): img with shape (H, W, C)
            label (Optional[np.ndarray]): label with shape (H, W)

        Returns: Scaled img and label(if exit)
        """
        scale_factor = self.scale_factors[self.output_index]

        return scale_img_label(scale_factor, img, label, self.aligner,
                               align_corner=self.align_corner, PIL_mode=self.PIL_mode)


def pad_img_label(img: np.ndarray, label: Optional[np.ndarray] = None, pad_img_to: Union[Iterable, int] = 0,
                  pad_aligner: Union[Callable[[int], int], Iterable[Callable[[int], int]]] = lambda x: x,
                  img_pad_val: Union[int, float, Iterable] = 0.0, ignore_label: int = 255,
                  pad_location: Union[str, int] = 'right-bottom',
                  with_mask: bool=False) \
        -> Union[np.ndarray, Tuple[np.ndarray, np.ndarray], Tuple[np.ndarray, np.ndarray,
                                                                  np.ndarray, Tuple[int, int, int, int]]]:
    """Pad img to size pad_aligner(max(img_origin_size, pad_img_to))

    Args:
        img (np.ndarray): img to be padded
        label (np.ndarray): label to be padded
        pad_img_to (Union[None, Iterable, int]): img pad size. If value is int, the img_pad_to will be parsed as
            H=value, W=value. Else will be parsed as H=list(value)[0], W=list(value)[1]
        pad_aligner (Union[Callable, Iterable[Callable]]): Final pad size will be refined by callable aligner. If
            Iterable, then first and second aligners separately used to align H and W.
        img_pad_val: (Union[int, float, Iterable]): If value is int or float, return (value, value, value),
            if value is Iterable with 3 element, return totuple(value), else raise error
        ignore_label (int): value to pad the label.
        pad_location (Union[str, int]): Indicate pad location. Can be 'left-top'/0, 'right-top'/1, 'left-bottom'/2
            'right-bottom'/3, 'center'/4.
        with_mask: If True, return mask of padded area and (i, j, h, w) of img in padded img.

    Returns: Padded img and label(if given)
    """
    # * Check img size = label size
    _check_img_size_equal_label_size(img, label)

    # * Check ignore label value
    if not is_int(ignore_label):
        raise ValueError(f"ignore_label{ignore_label} should be int")
    else:
        ignore_label = int(ignore_label)

    # * Get img pad scalar. If img is int array, then pad val can't be float.
    img_pad_scalar = color2scalar(img_pad_val)
    if is_intarr(img) and is_floatarr(np.array(img_pad_val)):
        raise ValueError("Input img is int array, while img_pad_val has float. Which may cause unexpected "
                         "behaviour when padding. Using int_img2float32img before padding int img with float value.")

    # * Get pad aligner
    if isinstance(pad_aligner, abc.Callable):
        pad_aligner_h, pad_aligner_w = pad_aligner, pad_aligner
    elif isinstance(pad_aligner, abc.Iterable):
        pad_aligner_h, pad_aligner_w = pad_aligner
    else:
        raise ValueError(f"pad_aligner {pad_aligner} must be Callable or Iterable[Callable]")

    # * Get pad_h, pad_w
    pad_to_h, pad_to_w = size2HW(pad_img_to)
    img_h, img_w = size2HW(img.shape[:2])
    padded_h, padded_w = pad_aligner_h(max(pad_to_h, img_h)), pad_aligner_w(max(pad_to_w, img_w))
    pad_h, pad_w = padded_h - img_h, padded_w - img_w

    if pad_h < 0 or pad_w < 0:
        raise ValueError("pad aligner's return size must larger than input size")

    # * Get right/left/bottom/top pad size
    if isinstance(pad_location, str):
        pad_location_index_dict = {'left-top': 0, 'right-top': 1, 'left-bottom': 2, 'right-bottom': 3, 'center': 4}
        pad_location_index = pad_location_index_dict.get(pad_location)
        if pad_location is None:
            raise ValueError(f"pad_location should be in {list(pad_location_index_dict.keys())}")
    else:
        pad_location_index = pad_location
        if pad_location_index not in [0, 1, 2, 3, 4]:
            raise ValueError(f"If pad location is given by int, then it should be in range 0~4")

    pad_left, pad_right, pad_top, pad_bottom = 0, 0, 0, 0
    if pad_location_index & 1:
        pad_right = pad_w
    else:
        pad_left = pad_w
    if pad_location_index & 2:
        pad_bottom = pad_h
    else:
        pad_top = pad_h
    if pad_location_index & 4:
        pad_top, pad_bottom = pad_h // 2, (pad_h + 1) // 2  # floor(pad_h/2), ceil(pad_h/2)
        pad_left, pad_right = pad_w // 2, (pad_w + 1) // 2

    ret = []

    img = cv2.copyMakeBorder(img, pad_top, pad_bottom, pad_left, pad_right,
                             borderType=cv2.BORDER_CONSTANT, value=img_pad_scalar)
    ret.append(img)

    if label is not None:
        label = cv2.copyMakeBorder(label, pad_top, pad_bottom, pad_left, pad_right,
                                   borderType=cv2.BORDER_CONSTANT, value=ignore_label)
        ret.append(label)

    if with_mask:
        mask = np.ones_like(img[:, :, 0], dtype=bool)
        mask[pad_top:pad_top + img_h, pad_left:pad_left + img_w] = False
        ret.append(mask)
        ret.append(np.asarray((pad_top, pad_left, img_h, img_w)))  # (i, j, h, w)

    return ret[0] if len(ret) == 1 else tuple(ret)


def int_img2float32_img(img: np.ndarray, scale: bool=False) -> np.ndarray:
    """Convert int img to float32 img

    Args:
        img (np.ndarray): int numpy img
        scale (bool): If True, img will be scale to range [0.0, 1.0]. (Default: False)

    Returns: float32 img
    """
    if not is_intarr(img):
        raise ValueError(f"img {img} is supposed to be an int arr")
    img = img.astype(np.float32)  # Return copy
    if scale:
        img /= 255  # Can be inplace
    return img


def centralize(img: np.ndarray, mean: Union[int, float, Iterable], std: Union[int, float, Iterable, None] = None):
    """Centralize img by minus the mean and divide the std

    Args:
        img: float img with size (H, W, C) to be centralized
        mean: mean value to be minus
        std: If not None, std value to be divided after img is minuted mean value

    Returns: Centralized img
    """
    if not is_floatarr(img):
        raise ValueError(f"img to be centralized should be float array")

    mean_scalar, std_scalar = np.array(color2scalar(mean), dtype=np.float32), \
        np.array(color2scalar(std), dtype=np.float32) if std is not None else None

    img = img - mean  # Return copy
    if std_scalar is not None:
        img /= std  # Inplace

    return img


def _crop_img_label(img, label, offset_h, offset_w, crop_h, crop_w) -> Union[np.ndarray, Tuple[np.ndarray]]:
    cropped_img = img[offset_h:offset_h + crop_h, offset_w:offset_w + crop_w, :]
    if label is None:
        return cropped_img
    else:
        cropped_label = label[offset_h:offset_h + crop_h, offset_w:offset_w + crop_w]
        return cropped_img, cropped_label


class RandCrop(RandMap):

    def __init__(self, crop_size: Union[int, Iterable[int]]):
        """Random crop img to crop_size

        Args:
            crop_size: Crop size. If size is int, the crop_height=value, crop_width=value. Else will be parsed as
                crop_height=list(value)[0], crop_width=list(value)[1]
        """
        super(RandCrop, self).__init__()

        self.crop_h, self.crop_w = size2HW(crop_size)

    def generate_rand_seed(self, img: np.ndarray, label: Optional[np.ndarray] = None):
        img_h, img_w = img.shape[:2]

        if img_h < self.crop_h or img_w < self.crop_w:
            raise ValueError(f"img_h {img_h} must >= crop_h {self.crop_h}; img_w {img_w} must >= crop_w {self.crop_w}")

        offset_h = np.random.randint(0, img_h - self.crop_h + 1)
        offset_w = np.random.randint(0, img_w - self.crop_w + 1)

        return int(offset_h), int(offset_w)

    def forward(self, img: np.ndarray, label: Optional[np.ndarray] = None) -> Union[np.ndarray, Tuple[np.ndarray]]:
        """Generate rand cropped img and label accroding to rand seed

        Args:
            img: img to be cropped
            label: If not None, means label to be cropped

        Returns: Cropped img and label(if not None)
        """
        _check_img_size_equal_label_size(img, label)

        img_h, img_w = img.shape[:2]
        offset_h, offset_w = self.rand_seed
        if offset_h + self.crop_h > img_h or offset_w + self.crop_w > img_w:
            raise ValueError(f"Crop size out of range. offset_h + crop_h = {offset_h + self.crop_h}, "
                             f"offset_w + crop_w = {offset_w + self.crop_w}, while img has shape {(img_h, img_w)}")

        return _crop_img_label(img, label, offset_h, offset_w, self.crop_h, self.crop_w)


class FiveCrop(MultiMap):
    output_num = 5

    def __init__(self, crop_size: Union[int, Iterable[int]]):
        """Five crop img to crop_size, with index 0-4 meaning left-top, right-top, left-bottom, right-bottom, center

        Args:
            crop_size: Crop size. If size is int, the crop_height=value, crop_width=value. Else will be parsed as
                crop_height=list(value)[0], crop_width=list(value)[1]
        """
        super(FiveCrop, self).__init__()

        self.crop_h, self.crop_w = size2HW(crop_size)

    def forward(self, img: np.ndarray, label: Optional[np.ndarray] = None) -> Union[np.ndarray, Tuple[np.ndarray]]:
        """Return cropped img and label according to output_index

        Args:
            img: img to be cropped
            label: If not None, label to be cropped

        Returns: Cropped img and label(If exits)
        """
        _check_img_size_equal_label_size(img, label)

        img_h, img_w = img.shape[:2]
        if img_h < self.crop_h or img_w < self.crop_w:
            raise ValueError(f"img_h {img_h} must >= crop_h {self.crop_h}; img_w {img_w} must >= crop_w {self.crop_w}")

        bias_h, bias_w = img_h - self.crop_h, img_w - self.crop_w
        offset_h, offset_w = 0, 0

        if self.output_index & 1:
            offset_w += bias_w
        if self.output_index & 2:
            offset_h += bias_h
        if self.output_index & 4:
            offset_h, offset_w = bias_h // 2, bias_w // 2

        return _crop_img_label(img, label, offset_h, offset_w, self.crop_h, self.crop_w)


class RandRangeScale(RandMap):

    def __init__(self, low_size: int, high_size: int, short_thresh: int,
                 aligner: Union[Callable[[int], int], Iterable[Callable[[int], int]]] = lambda x: x,
                 align_corner: bool=True, PIL_mode: int | None=None):
        """在保证短边大于阈值的情况下，尝试将长边缩放到随机尺寸。

        Args:
            low_size: 长边随机缩放的最短尺寸。
            high_size: 长边随机缩放的最大尺寸。
            short_thresh: 短边缩放后的最小尺寸。
            align_corner: 缩放时是否对齐角点。
        """
        super(RandRangeScale, self).__init__()

        self.low_size = low_size
        self.high_size = high_size
        self.short_thresh = short_thresh
        self.aligner = aligner
        self.align_corner = align_corner
        self.PIL_mode = PIL_mode

    def generate_rand_seed(self, img: np.ndarray, label: Optional[np.ndarray] = None) -> float:
        # * 先计算一个“目标尺寸”。
        object_size = self.low_size + np.random.randint(self.high_size - self.low_size + 1)

        img_h, img_w = img.shape[:2]

        long_size = img_h if img_h > img_w else img_w
        short_size = img_h if img_h < img_w else img_w
        # * 若长边达到目标尺寸的同时，短边大于short thresh，则让长边缩放到目标尺寸。否则将短边缩放到阈值尺寸。
        scale_ratio = max(object_size / long_size, self.short_thresh / short_size)
        return scale_ratio

    def forward(self, img: np.ndarray, label: Optional[np.ndarray] = None) -> Union[np.ndarray, Tuple[np.ndarray]]:
        return scale_img_label(self.rand_seed, img, label,
                               aligner=self.aligner, align_corner=self.align_corner, PIL_mode=self.PIL_mode)


class RandomResizeCrop(RandMap):
    """随机裁剪图像。"""

    def __init__(
            self,
            size: Union[int, Iterable[int]],
            scale: tuple[float, float]=(0.08, 1.0),
            ratio: tuple[float, float]=(3.0 / 4.0, 4.0 / 3.0),
            interpolation=Image.BILINEAR):
        """封装torchvision的RandomResizedCrop，实现对image和label的同步增强。

        Args:
            size: 裁剪后的尺寸。
            scale: 随机缩放的范围。
            ratio: 随机裁剪的宽高比范围。
            interpolation: 插值方式。
        """
        super(RandomResizeCrop, self).__init__()

        self.transform = T.RandomResizedCrop(size, scale, ratio, interpolation)

        self.size = size
        self.scale = scale
        self.ratio = ratio
        self.interpolation = interpolation

    def generate_rand_seed(self, img: np.ndarray, label: Optional[np.ndarray] = None) -> tuple[int, int, int, int]:
        transform = self.transform
        pil_img = arr2PIL(img, order='RGB')  # 只用于获取尺寸，通道顺序无所谓。
        return transform.get_params(pil_img, transform.scale, transform.ratio)

    def forward(self, img: np.ndarray, label: Optional[np.ndarray] = None) -> Union[np.ndarray, Tuple[np.ndarray]]:
        transform = self.transform
        i, j, h, w = self.rand_seed

        pil_img = arr2PIL(img, order='RGB')  # 只用于缩放，通道顺序无所谓。
        scaled_img = TF.resized_crop(pil_img, i, j, h, w, transform.size, transform.interpolation)

        if label is not None:
            pil_label = arr2PIL(label)
            scaled_label = TF.resized_crop(pil_label, i, j, h, w, transform.size, Image.NEAREST)
            return PIL2arr(scaled_img, order='RGB'), PIL2arr(scaled_label)
        else:
            return PIL2arr(scaled_img, order='RGB')


def identical(x):
    """原样返回单个参数。"""
    return x


def pack_identical(*x):
    """原样返回多个参数。"""
    return x


def arr2PIL(arr: np.ndarray, order: Literal['RGB', 'BGR']='BGR') -> Image.Image:
    """将numpy数组转为PIL图片。

    Args:
        arr: (H, W, [C])的numpy数组。
        order: 彩图的通道顺序。对单通道图无效。

    Returns:
        PIL图片。
    """
    is_color = arr.ndim == 3
    if is_color and order == 'BGR':
        arr = BGR2RGB(arr)
    return Image.fromarray(arr, mode='RGB' if is_color else 'L')


def PIL2arr(pil: Image.Image, order: Literal['RGB', 'BGR']='BGR') -> np.ndarray:
    """将numpy数组转为PIL图片。

    Args:
        pil: PIL图片。
        order: 彩图的通道顺序。对单通道图无效。

    Returns:
        numpy数组。
    """
    arr = np.asarray(pil)
    if pil.mode == 'L':
        return arr
    elif pil.mode == 'RGB':
        if order == 'RGB':
            return arr
        elif order == 'BGR':
            return RGB2BGR(arr)
        else:
            raise ValueError(f"不支持的{order=}。")
    else:
        raise ValueError(f"不支持的PIL模式{pil.mode=}。")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author  : Xiaobo Yang
@Contact : hal_42@zju.edu.cn
@Time    : 2022/9/14 15:24
@File    : swa_train_env.py
@Software: PyCharm
@Desc    : 
"""
import torch
import torch.nn as nn

from alchemy_cat.py_tools import set_torch_rand_seed

from ..rng_cacher import RNGCacher

__all__ = ['SWATrainEnv']


class SWATrainEnv(object):
    """创建SWA训练环境。"""
    def __init__(self, model: nn.Module, device: str | int | torch.device, seed: int | str,
                 norm_init: bool=True, cumulative_avg=True):
        """SWA训练环境。

        Args:
            model: SWA模型。
            device: SWA模型训练时所在设备。
            seed: SWA训练前设置的随机种子。
            norm_init: SWA模型训练前是否用标准初始化BN统计量。
            cumulative_avg: 为True，SWA模型训练时用累积法计算统计量，反之则动量更新统计量。
        """
        self.model = model
        self.device = device
        self.seed = seed
        self.norm_init = norm_init
        self.cumulative_avg = cumulative_avg

        self.out_grad_enabled: bool | None = None
        self.cacher = RNGCacher()
        self.out_model_device: bool | None = None
        self.out_model_training: bool | None = None
        self.momenta: dict[nn.Module, ...] = {}

    def __enter__(self) -> nn.Module:
        # * 记录外部梯度可用性，关闭梯度。
        self.out_grad_enabled = torch.is_grad_enabled()
        torch.set_grad_enabled(False)
        # * 记录外部随机性。
        self.cacher.cache()
        # * 重设随机种子。
        set_torch_rand_seed(self.seed)
        # * 记录外部模型所在设备，并迁移模型到指定设备。
        self.out_model_device = next(self.model.parameters()).device
        self.model.to(device=self.device)
        # * 记录外部模型是否可训，并设置为训练状态。
        self.out_model_training = self.model.training
        self.model.train()
        # * 记录模型动量。
        for module in self.model.modules():
            if isinstance(module, nn.modules.batchnorm._BatchNorm):
                self.momenta[module] = module.momentum
                if self.norm_init:
                    module.running_mean = torch.zeros_like(module.running_mean)
                    module.running_var = torch.ones_like(module.running_var)
                if self.cumulative_avg:
                    module.momentum = None
                    module.num_batches_tracked *= 0
        assert len(self.momenta) > 0
        return self.model

    def __exit__(self, exc_type, exc_val, exc_tb):
        # * 恢复BN层动量。
        for bn_module in self.momenta.keys():
            bn_module.momentum = self.momenta[bn_module]
        # * 恢复外部模型可训性。
        self.model.train(self.out_model_training)
        # * 模型迁移到在外部时的设备。
        self.model.to(device=self.out_model_device)
        # * 恢复外部随机状态。
        self.cacher.resume(purge=True)
        # * 恢复外部梯度可用性。
        torch.set_grad_enabled(self.out_grad_enabled)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author  : Xiaobo Yang
@Contact : hal_42@zju.edu.cn
@Time    : 2022/5/25 21:49
@File    : cfg.py
@Software: PyCharm
@Desc    : 
"""
from alchemy_cat.py_tools import Cfg2Tune


cfg = config = Cfg2Tune('configs/base_cfg/cfg.py', 'configs/update_cfg/cfg.py')

cfg.rslt_dir = 'init_with_update'



#!/usr/bin/env src
# -*- coding:utf-8 -*-
#
# Author:  yaoqi (yaoqi_isee@zju.edu.cn)
# Created Date: 2019-12-25
# Modified By: yaoqi (yaoqi_isee@zju.edu.cn)
# Last Modified: 2019-12-30
# -----
# Copyright (c) 2019 Zhejiang University
import torch
import torch.nn as nn
import torch.nn.functional as F

from alchemy_cat.torch_tools import Network

__all__ = ['PAM']


class PAM(Network):
    def __init__(self, in_channel, mid_channel, num_classes, layer_idx, branch_idx):
        """init function for positional attention module
        
        Parameters
        ----------
        in_channel : int    
            channel of input tensor
        mid_channel : int
            channel of temporal variables
        num_classes : int
            number of classes
        """
        super(PAM, self).__init__()

        self.in_channel = in_channel
        self.out_channel = num_classes

        self.in_dim, self.out_dim = in_channel, in_channel
        self.key_dim, self.query_dim, self.value_dim = mid_channel, mid_channel, in_channel
        self.f_key = nn.Conv2d(self.in_dim, self.key_dim, 1)
        self.f_query = nn.Conv2d(self.in_dim, self.query_dim, 1)

        self.gamma = nn.Parameter(torch.zeros(1), requires_grad=True)
        # self.fc = nn.Conv2d(self.in_dim, num_classes, 3, padding=12, dilation=12)
        self.layer_idx = layer_idx
        self.branch_idx = branch_idx
        self.add_module(f"fc{layer_idx}_{branch_idx}", nn.Conv2d(self.in_dim, num_classes, 1))

    def forward(self, x):
        batch_size, h, w = x.size(0), x.size(2), x.size(3)
        value = x.view(batch_size, self.value_dim, -1)
        value = value.permute(0, 2, 1)

        query = self.f_query(x).view(batch_size, self.query_dim, -1)
        query = query.permute(0, 2, 1)

        key = self.f_key(x).view(batch_size, self.key_dim, -1)

        sim_map = torch.matmul(query, key)
        sim_map = (self.key_dim ** -.5) * sim_map
    
        sim_map = F.softmax(sim_map, dim=-1)

        context = torch.matmul(sim_map, value)
        context = context.permute(0, 2, 1).contiguous()
        context = context.view(batch_size, self.value_dim, *x.size()[2:])

        fuse = self.gamma * context + x

        score = getattr(self, f"fc{self.layer_idx}_{self.branch_idx}")(fuse)
        return score



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author  : Xiaobo Yang
@Contact : hal_42@zju.edu.cn
@Time    : 2022/8/2 16:28
@File    : softmax.py
@Software: PyCharm
@Desc    :
"""
import torch
import torch.nn as nn
import torch.nn.functional as F


__all__ = ['MaskedSoftmax']


class MaskedSoftmax(nn.Module):

    def __init__(self, dim: int=-1, epsilon: float=1e-5):
        super(MaskedSoftmax, self).__init__()
        self.dim = dim
        self.epsilon = epsilon

        self.softmax = nn.Softmax(dim=dim)

    def forward(self, sim_map: torch.Tensor, mask: torch.Tensor | None=None) -> torch.Tensor:
        """做有掩码的Softmax。

        Args:
            sim_map: logits。
            mask: 与logits同shape的掩码。

        Returns:
            带掩码的Softmax结果。
        """
        if mask is None:
            return F.softmax(sim_map, dim=self.dim)
        else:
            e_sim = torch.exp(sim_map - torch.max(sim_map, dim=self.dim, keepdim=True)[0])
            masked_e_sim = e_sim * mask
            masked_sums = masked_e_sim.sum(dim=self.dim, keepdim=True) + self.epsilon
            return masked_e_sim / masked_sums



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author  : Xiaobo Yang
@Contact : hal_42@zju.edu.cn
@Time    : 2021/7/30 21:22
@File    : py3_send_recv.py
@Software: PyCharm
@Desc    :
"""
from typing import Any

import sys
import atexit
import os
import os.path as osp
import shutil
import pickle
import signal
import glob
from time import sleep

import numpy as np

__all__ = ['client_send', 'server_send', 'client_recv', 'server_recv', 'init_server']


kBufferDirPrefix = 'Temp' if os.name == 'nt' else osp.join('/dev', 'shm')
kFileName = '0'

kSleepTime = .0001  # 0.1ms

kClientBufferDirSuffix = 'client_buffer'
kServerBufferDirSuffix = 'server_buffer'

kLockFile = osp.join(kBufferDirPrefix, 'server.lock')


def _save_obj_sent(obj_file: str, obj: Any) -> str:
    if isinstance(obj, np.ndarray):
        obj_file = obj_file + '.npy'
        np.save(obj_file, obj)
    else:
        obj_file = obj_file + '.pkl'
        with open(obj_file, 'wb') as pkl_f:
            pickle.dump(obj, pkl_f, protocol=2)

    return obj_file


def _load_obj_recv(obj_file: str) -> Any:
    file_type = osp.splitext(obj_file)[-1]
    if file_type == '.npy':
        obj = np.load(obj_file)
    elif file_type == '.pkl':
        with open(obj_file, 'rb') as pkl_f:
            obj = pickle.load(pkl_f)
    else:
        raise RuntimeError("Receive unsupported file type %s" % file_type)

    return obj


def shm_send(obj: Any, buffer_dir_suffix: str, time_out: int=-1) -> str:
    """Send object to buffer in share memory.

    Args:
        obj: object to be send.
        buffer_dir_suffix: buffer dir name.
        time_out: Max time to wait, 0.1ms as unit. If time_out < 0, then can wait for any longer. (Default: -1)

    Returns: Object send path.
    """
    buffer_dir = osp.join(kBufferDirPrefix, buffer_dir_suffix)
    os.makedirs(buffer_dir, exist_ok=True)

    signal_file = osp.join(buffer_dir, kFileName + '.signal')
    obj_file = osp.join(buffer_dir, kFileName)

    wait_times = 0
    while osp.isfile(signal_file):
        sleep(kSleepTime)
        wait_times += 1
        if 0 <= time_out < wait_times:
            raise TimeoutError("Send option has waited for %d times" % wait_times)

    _save_obj_sent(obj_file, obj)
    with open(signal_file, mode='w'):
        pass

    return obj_file


def shm_recv(buffer_dir_suffix: str, time_out: int=-1) -> Any:
    """Receive object from buffer in share memory.

    Args:
        buffer_dir_suffix: buffer dir name.
        time_out: Max time to wait, 0.1ms as unit. If time_out < 0, then can wait for any longer. (Default: -1)

    Returns: Received object.
    """
    buffer_dir = osp.join(kBufferDirPrefix, buffer_dir_suffix)
    signal_file = osp.join(buffer_dir, kFileName + '.signal')

    wait_times = 0
    while not osp.isfile(signal_file):
        sleep(kSleepTime)
        wait_times += 1
        if 0 <= time_out < wait_times:
            raise TimeoutError("Recv option has waited for %d times" % wait_times)

    ls = glob.glob(osp.join(buffer_dir, kFileName) + '.*')
    ls.remove(signal_file)
    assert len(ls) == 1
    obj_file = ls[0]

    obj = _load_obj_recv(obj_file)

    os.remove(obj_file)
    os.remove(signal_file)

    return obj


def client_send(obj: Any, time_out: int=-1) -> str:
    """Send object to buffer in client buffer.

    Args:
        obj: object to be send.
        time_out: Max time to wait, 0.1ms as unit. If time_out < 0, then can wait for any longer. (Default: -1)

    Returns: Img send path.
    """
    return shm_send(obj, buffer_dir_suffix=kClientBufferDirSuffix, time_out=time_out)


def server_send(obj: Any, time_out: int=-1) -> str:
    """Send object to buffer in server buffer.

    Args:
        obj: object to be send.
        time_out: Max time to wait, 0.1ms as unit. If time_out < 0, then can wait for any longer. (Default: -1)

    Returns: Img send path.
    """
    return shm_send(obj, buffer_dir_suffix=kServerBufferDirSuffix, time_out=time_out)


def client_recv(time_out: int=-1) -> Any:
    """Receive object from server buffer.

    Args:
        time_out: Max time to wait, 0.1ms as unit. If time_out < 0, then can wait for any longer. (Default: -1)

    Returns: Received object.
    """
    return shm_recv(kServerBufferDirSuffix, time_out=time_out)


def server_recv(time_out: int=-1) -> Any:
    """Receive object from client buffer.

    Args:
        time_out: Max time to wait, 0.1ms as unit. If time_out < 0, then can wait for any longer. (Default: -1)

    Returns: Received object.
    """
    return shm_recv(kClientBufferDirSuffix, time_out=time_out)


def init_server():
    """Init server
    This function will:
        1. Make sure only one server are running.
        2. Show welcome word.
        3. Make sure the buffer will be removed after server is terminated.
        4. Create buffer dir.
    """
    if osp.isfile(kLockFile):
        raise RuntimeError("Server has already been ran.")

    with open(kLockFile, "w"):
        pass

    print("Server Start on share memory at %s" % kBufferDirPrefix)

    def clear_buffer():
        if osp.isdir(osp.join(kBufferDirPrefix, kClientBufferDirSuffix)):
            shutil.rmtree(osp.join(kBufferDirPrefix, kClientBufferDirSuffix))
        if osp.isdir(osp.join(kBufferDirPrefix, kServerBufferDirSuffix)):
            shutil.rmtree(osp.join(kBufferDirPrefix, kServerBufferDirSuffix))
        if osp.isfile(kLockFile):
            # * Only one process need to clean the Lock File.
            os.remove(kLockFile)

    def clear_exit(signum, _):
        print("Get signal: %s, clearing buffer..." % signum)
        clear_buffer()
        print("Exit server. ")
        sys.exit(0)

    atexit.register(clear_buffer)
    signal.signal(signal.SIGHUP, clear_exit)
    signal.signal(signal.SIGINT, clear_exit)
    signal.signal(signal.SIGTERM, clear_exit)

    os.makedirs(osp.join(kBufferDirPrefix, kClientBufferDirSuffix), exist_ok=True)
    os.makedirs(osp.join(kBufferDirPrefix, kServerBufferDirSuffix), exist_ok=True)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author  : Xiaobo Yang
@Contact : hal_42@zju.edu.cn
@Time    : 2024/6/24 19:12
@File    : test_adict.py
@Software: PyCharm
@Desc    : 
"""
import pickle
from copy import deepcopy

import pytest

from alchemy_cat.py_tools import ADict

__all__ = ['d', 'ref_d', 'dic', 'ref_dic', 'to_update', 'dic_can_exec']


@pytest.fixture(scope="function")  # 字典可能被修改，所以每次都要重新生成。
def d() -> dict:
    yield {'num': 0,
           'lst': [1, [2, {'lst_lst_dic': 'exit'}], {'lst_dic': 'exit'}],
           'dic': {'sub_num': 3, 'sub_set': frozenset((1, 2, 3))},
           'uni_dic': {'sub_num': 4},
           'obj': object()}


@pytest.fixture(scope="function")
def ref_d(d) -> dict:
    d['lst'][1].append(d['lst'])
    d['inner_lst'] = d['lst'][1]
    d['lst'].append(d)
    d['uni_dic']['prev_dic'] = d['dic']
    yield d


@pytest.fixture(scope="function")
def dic(d) -> ADict:
    yield ADict(d)


@pytest.fixture(scope="function")
def ref_dic(ref_d) -> ADict:
    yield ADict(ref_d)

# -* 测试拷贝时自引用。


def test_ref_dic(ref_dic: ADict):
    assert ref_dic['lst'][1][-1] is ref_dic['lst']
    assert ref_dic['inner_lst'] is ref_dic['lst'][1]
    assert ref_dic['lst'][-1] is ref_dic
    assert ref_dic['uni_dic']['prev_dic'] is ref_dic['dic']


def test_ref_dic_copy(ref_dic: ADict):
    copy_ref_dic = ref_dic.copy()
    assert copy_ref_dic['lst'][1][-1] is copy_ref_dic['lst']
    assert copy_ref_dic['inner_lst'] is copy_ref_dic['lst'][1]
    assert copy_ref_dic['lst'][-1] is copy_ref_dic
    assert copy_ref_dic['uni_dic']['prev_dic'] is copy_ref_dic['dic']


def test_deepcopy_inner_outer_ref(ref_dic: ADict):
    ref_dic_lst = [ref_dic, ref_dic, ref_dic]
    nl = deepcopy(ref_dic_lst)
    assert nl[0] is nl[1] is nl[2] is not ref_dic_lst[0]
    assert nl[0]['lst'][1][-1] is nl[0]['lst']
    assert nl[0]['inner_lst'] is nl[0]['lst'][1]
    assert nl[0]['lst'][-1] is nl[0]
    assert nl[0]['uni_dic']['prev_dic'] is nl[0]['dic']


def test_pickle_ref_dic(ref_dic: ADict):
    nd = pickle.loads(pickle.dumps(ref_dic))
    assert nd['lst'][1][-1] is nd['lst']
    assert nd['inner_lst'] is nd['lst'][1]
    assert nd['lst'][-1] is nd
    assert nd['uni_dic']['prev_dic'] is nd['dic']


def test_todict_ref(ref_dic: ADict):
    to_dic = ref_dic.to_dict()
    assert to_dic['lst'][1][-1] is to_dic['lst']
    assert to_dic['inner_lst'] is to_dic['lst'][1]
    assert to_dic['lst'][-1] is to_dic
    assert to_dic['uni_dic']['prev_dic'] is to_dic['dic']


def test_pickle_remain_whole_frozen(dic: ADict):
    dic.uni_dic.set_whole()
    dic.freeze()
    nd = pickle.loads(pickle.dumps(dic))
    assert nd.uni_dic.is_whole
    assert nd.is_frozen
    assert nd.dic.is_frozen
    assert nd.uni_dic.is_frozen


def test_branches_dict_ref(ref_dic: ADict):
    b = list(ref_dic.branches)
    assert len(b) == 3
    assert b[0] is ref_dic
    assert b[1] is ref_dic.dic
    assert b[2] is ref_dic.uni_dic


def test_ckl_dict_ref(ref_dic: ADict):
    ns, cs, ks, ls = [], [], [], []
    for name, (c, k, l) in ref_dic.named_ckl:
        ns.append(name)
        cs.append(c)
        ks.append(k)
        ls.append(l)

    assert len(ns) == len(cs) == len(ks) == len(ls) == 7

    assert ls[0] == 0
    assert ls[1] is ref_dic.lst
    assert ls[2] == 3
    assert ls[3] is ref_dic.dic.sub_set
    assert ls[4] == 4
    assert ls[5] is ref_dic.obj
    assert ls[6] is ref_dic.inner_lst

    assert ns[0] == 'num'
    assert ns[1] == 'lst'
    assert ns[2] == 'dic.sub_num'
    assert ns[3] == 'dic.sub_set'
    assert ns[4] == 'uni_dic.sub_num'
    assert ns[5] == 'obj'
    assert ns[6] == 'inner_lst'

    assert cs[0] is ref_dic
    assert cs[1] is ref_dic
    assert cs[2] is ref_dic.dic
    assert cs[3] is ref_dic.dic
    assert cs[4] is ref_dic.uni_dic
    assert cs[5] is ref_dic
    assert cs[6] is ref_dic

    assert ks[0] == 'num'
    assert ks[1] == 'lst'
    assert ks[2] == 'sub_num'
    assert ks[3] == 'sub_set'
    assert ks[4] == 'sub_num'
    assert ks[5] == 'obj'
    assert ks[6] == 'inner_lst'


# -* 测试dict_update。


@pytest.fixture(scope="function")
def to_update() -> ADict:
    to_update = ADict()
    to_update.my_num = -1
    to_update.lst = 'lst'  # 人有我有，都是value。
    to_update.dic.sub_num = 30  # 人有我有，都是子树 -> 人有我有，都是value。
    to_update.dic.sub_tuple = (10,)  # 人有我有，都是子树 -> 人无我有。
    yield to_update


def test_update(to_update: ADict, dic: ADict):
    to_update.update(dic.freeze(), dic={'add_by_kwargs': 8}, new_key='lala')
    assert to_update == {'my_num': -1,
                         'lst': [1, [2, {'lst_lst_dic': 'exit'}], {'lst_dic': 'exit'}],
                         'dic': {'sub_num': 3, 'sub_tuple': (10,), 'sub_set': frozenset({1, 2, 3}), 'add_by_kwargs': 8},
                         'num': 0,
                         'uni_dic': {'sub_num': 4},
                         'obj': to_update.obj,
                         'new_key': 'lala'}
    assert to_update.obj is dic.obj
    assert to_update.lst is not dic.lst
    assert to_update.uni_dic is not dic.uni_dic


def test_update_whole(to_update: ADict, dic: ADict):
    dic = dic.freeze()
    dic.dic.set_whole()
    to_update.update(dic)
    assert to_update == {'my_num': -1,
                         'lst': [1, [2, {'lst_lst_dic': 'exit'}], {'lst_dic': 'exit'}],
                         'dic': {'sub_num': 3, 'sub_set': frozenset((1, 2, 3))},
                         'num': 0,
                         'uni_dic': {'sub_num': 4},
                         'obj': to_update.obj}


def test_update_incremental(to_update: ADict, dic: ADict):
    to_update.dict_update(dic.freeze(), incremental=True)
    assert to_update == {'my_num': -1,
                         'lst': 'lst',
                         'dic': {'sub_num': 30, 'sub_tuple': (10,), 'sub_set': frozenset((1, 2, 3))},
                         'num': 0,
                         'uni_dic': {'sub_num': 4},
                         'obj': to_update.obj}


def test_update_or(to_update: ADict, dic: ADict):
    new_dic = to_update | dic.freeze()
    assert new_dic == {'my_num': -1,
                       'lst': [1, [2, {'lst_lst_dic': 'exit'}], {'lst_dic': 'exit'}],
                       'dic': {'sub_num': 3, 'sub_tuple': (10,), 'sub_set': frozenset({1, 2, 3})},
                       'num': 0,
                       'uni_dic': {'sub_num': 4},
                       'obj': dic.obj}
    assert new_dic is not to_update


def test_update_ror(to_update: ADict, dic: ADict):
    new_dic = dic.to_dict() | to_update
    assert new_dic == {'my_num': -1,
                       'lst': [1, [2, {'lst_lst_dic': 'exit'}], {'lst_dic': 'exit'}],
                       'dic': {'sub_num': 3, 'sub_tuple': (10,), 'sub_set': frozenset({1, 2, 3})},
                       'num': 0,
                       'uni_dic': {'sub_num': 4},
                       'obj': dic.obj}
    assert new_dic is not to_update


def test_update_ior(to_update: ADict, dic: ADict):
    ori_id = id(to_update)
    to_update |= dic.freeze()
    assert to_update == {'my_num': -1,
                         'lst': [1, [2, {'lst_lst_dic': 'exit'}], {'lst_dic': 'exit'}],
                         'dic': {'sub_num': 3, 'sub_tuple': (10,), 'sub_set': frozenset({1, 2, 3})},
                         'num': 0,
                         'uni_dic': {'sub_num': 4},
                         'obj': dic.obj}
    assert id(to_update) == ori_id

# -* 测试freeze功能。


def test_freeze(dic: ADict):
    dic.freeze()
    assert dic.is_frozen
    assert dic.dic.is_frozen
    assert dic.uni_dic.is_frozen

    with pytest.raises(RuntimeError):
        dic.new_key = 1

    with pytest.raises(KeyError):
        _ = dic.new_key

    with pytest.raises(RuntimeError):
        del dic.dic.sub_num

    with pytest.raises(RuntimeError):
        dic.clear()

    with pytest.raises(RuntimeError):
        dic.pop('new_key')

    with pytest.raises(RuntimeError):
        dic.popitem()

    with pytest.raises(RuntimeError):
        dic.dic.empty_leaf()

    dic.unfreeze()
    assert not dic.is_frozen
    assert not dic.dic.is_frozen
    assert not dic.uni_dic.is_frozen


# -* 测试重写的受影响的字典方法。

def test_dic_from_keys(dic: ADict):
    value = [99, 88, 77]
    dic_from_keys = ADict.fromkeys(['a', 'b', 'c'], [99, 88, 77])
    assert dic_from_keys == {'a': value, 'b': value, 'c': value}
    assert dic_from_keys['a'] is dic_from_keys['b'] is dic_from_keys['c']
    assert dic_from_keys['a'] is not value


# -* 测试set_whole。对其update的效果再update测试中已经完成。

def test_set_whole(dic: ADict):
    dic.dic.set_whole()
    assert dic.dic.is_whole

    dic.new.new_new.new_new_new.set_whole()
    assert 'new_new_new' in dic.new.new_new
    assert dic.new.new_new.new_new_new.is_whole

    dic.uni_dic.empty_leaf()
    assert dic.uni_dic.is_whole
    assert len(dic.uni_dic) == 0

    dic.dic.empty_tree()
    assert not dic.dic.is_whole
    assert len(dic.dic) == 0

    dic.new_empty.new_new_empty.empty_leaf()
    assert dic.new_empty.new_new_empty.is_whole
    assert len(dic.new_empty.new_new_empty) == 0

    dic.new_tree.new_new_tree.empty_tree()
    assert not dic.new_tree.new_new_tree.is_whole
    assert len(dic.new_tree.new_new_tree) == 0


# -* 测试to_txt()

@pytest.fixture(scope="function")
def dic_can_exec(dic: ADict) -> ADict:
    del dic.obj
    dic.uni_dic.set_whole()
    yield dic


def test_to_txt(dic_can_exec: ADict):
    exec(dic_can_exec.to_txt('exec_dic'), globals(), ldict := {})
    exec_dic = ldict['exec_dic']
    assert exec_dic == dic_can_exec
    assert exec_dic.uni_dic.is_whole



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@author: Xiaobo Yang
@contact: hal_42@zju.edu.cn
@software: PyCharm
@file: utils.py
@time: 2020/1/7 23:46
@desc:
"""
import numpy as np

from typing import Union, Iterable

from alchemy_cat.data.plugins.augers import pad_img_label

__all__ = ['stack_figs']


def stack_figs(in_list: list, img_pad_val: Union[int, float, Iterable] = (127, 140, 141),
               pad_location: Union[str, int]='right-bottom') -> np.ndarray:
    """Arrays in list will be padded to the max height/width in the list. The padding value will be 0 and located at
    right and bottom.

    Args:
        in_list (list): A list of numpy array shaped (H, W, C) with length N
        img_pad_val: (Union[int, float, Iterable]): If value is int or float, return (value, value, value),
            if value is Iterable with 3 element, return totuple(value), else raise error
        pad_location (Union[str, int]): Indicate pad location. Can be 'left-top'/0, 'right-top'/1, 'left-bottom'/2
            'right-bottom'/3, 'center'/4.

    Returns:
        np.ndarray: Stacked np array with shape (N, H, W, C)
    """
    max_h = 0
    max_w = 0
    for i, item in enumerate(in_list):
        if item.shape[0] > max_h:
            max_h = item.shape[0]
        if item.shape[1] > max_w:
            max_w = item.shape[1]
    out_arr = np.zeros((len(in_list), max_h, max_w, 3), dtype=in_list[0].dtype)
    for i, item in enumerate(in_list):
        out_arr[i] = pad_img_label(item, pad_img_to=(max_h, max_w), img_pad_val=img_pad_val, pad_location=pad_location)
    return out_arr



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author  : Xiaobo Yang
@Contact : hal_42@zju.edu.cn
@Time    : 2021/5/24 14:25
@File    : utils.py
@Software: PyCharm
@Desc    : 
"""
import numpy as np
from scipy.stats import entropy

__all__ = ["get_bg_fg_score", "cls_in_label2tag_always_bg", "cls_in_label2tag", "tag2cls_in_label",
           "complement_entropy_with_shift"]


def get_bg_fg_score(score: np.ndarray, alpha: float) -> np.ndarray:
    fg_score = score.reshape(score.shape[0], -1)
    bg_score = np.power(1 - np.amax(fg_score, axis=0, keepdims=True), alpha)
    bg_fg_score = np.concatenate((bg_score, fg_score), axis=0)
    return bg_fg_score.reshape((score.shape[0] + 1,) + score.shape[1:])


def cls_in_label2tag(cls_in_label: np.ndarray):
    return np.nonzero(cls_in_label)[0]


def cls_in_label2tag_always_bg(cls_in_label: np.ndarray):
    return np.insert(np.nonzero(cls_in_label[1:])[0] + 1, 0, 0, 0)


def tag2cls_in_label(tag: np.ndarray, num_cls: int):
    cls_in_label = np.zeros(num_cls, dtype=np.uint8)
    cls_in_label[tag] = 1
    return cls_in_label


def complement_entropy_with_shift(prob: np.ndarray, shift: float=0.1) -> np.ndarray:
    """Compute complement entropy with ( log2(cls_num) - entropy(prob) ) / log2(cls_num).

    Args:
        prob: (N, C, ...) prob. prob will be normalize if they don't sum to 1.
        shift: prob will be shifted by prob += shift. This will make only the diversity with the same magnitude as shift
            can influence the result.

    Returns:
        (N, ...) complement entropy.
    """
    return (np.log2(prob.shape[1]) - entropy(prob.swapaxes(0, 1) + shift, base=2)) / np.log2(prob.shape[1])



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author  : Xiaobo Yang
@Contact : hal_42@zju.edu.cn
@Time    : 2022/10/13 21:18
@File    : window_slides.py
@Software: PyCharm
@Desc    : 
"""
__all__ = ['slide_n_windows_len_l']


def slide_n_windows_len_l(n: int, l: int, lst: list | tuple) -> list:
    """从lst中截取尽量不相交、距离远的n个长l的切片。

    Args:
        n: 窗口数。
        l: 窗口长。
        lst: list序列。

    Returns:

    """
    step = (len(lst) - l) // (n - 1)

    return [lst[i * step:i * step + l] for i in range(n)]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author  : Xiaobo Yang
@Contact : hal_42@zju.edu.cn
@Time    : 2021/5/6 22:09
@File    : cfg.py.py
@Software: PyCharm
@Desc    : 
"""
from alchemy_cat.py_tools import Cfg2Tune, Param2Tune

cfg = config = Cfg2Tune()

cfg.rslt_dir = 'no_legal_val_config'

# * Param on same Cfg2Tune
cfg.foo0.foo1.a = Param2Tune([0, 1, 2])


def b_subject_to(opt_val):
    if cfg.foo0.foo1.a.cur_val > 1:
        return opt_val > 1
    return True


cfg.foo0.foo1.b = Param2Tune([.1, .2], b_subject_to)
# * Param on root Cfg2Tune


def c_subject_to(_):
    if cfg.foo0.foo1.b.cur_val == .2:
        return False
    return True


cfg.c = Param2Tune([False, True], c_subject_to)
# * Static Value
cfg.foo0.a_max = 2



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author  : Xiaobo Yang
@Contact : hal_42@zju.edu.cn
@Time    : 2021/4/15 23:07
@File    : __init__.py.py
@Software: PyCharm
@Desc    : 
"""
from alchemy_cat.torch_tools.network.network import *
from alchemy_cat.torch_tools.network.update_model_state_dict import *